
"""
This module includes tests for the L{xmantissa.scrolltable} module.
"""

from re import escape

from epsilon.hotfix import require

require("twisted", "trial_assertwarns")

from zope.interface import implements

from twisted.trial import unittest
from twisted.trial.util import suppress as SUPPRESS

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer, text
from axiom.dependency import installOn
from axiom.test.util import QueryCounter

from xmantissa.webapp import PrivateApplication
from xmantissa.ixmantissa import IWebTranslator, IColumn
from xmantissa.error import Unsortable


from xmantissa.scrolltable import (
    InequalityModel,
    ScrollableView,
    ScrollingFragment,
    ScrollingElement,
    SequenceScrollingFragment,
    StoreIDSequenceScrollingFragment,
    AttributeColumn,
    UnsortableColumnWrapper,
    UnsortableColumn)


_unsortableColumnSuppression = SUPPRESS(
    message=escape(
        "Use UnsortableColumnWrapper(AttributeColumn(*a, **kw)) "
        "instead of UnsortableColumn(*a, **kw)."),
    category=DeprecationWarning)



class DataThunk(Item):
    a = integer()
    b = integer()
    c = text()



class DataThunkWithIndex(Item):
    """
    Another testing utility, similar to L{DataThunk}, but with an indexed
    attribute, so that sorting complexity is independent of the number of
    instances which exist, so that performance testing can be done.
    """
    a = integer(indexed=True)



class ScrollTestMixin(object):
    def setUp(self):
        self.store = Store()
        installOn(PrivateApplication(store=self.store), self.store)
        self.six = DataThunk(a=6, b=8157, c=u'six', store=self.store)
        self.three = DataThunk(a=3, b=821375, c=u'three', store=self.store)
        self.seven = DataThunk(a=7, b=4724, c=u'seven', store=self.store)
        self.eight = DataThunk(a=8, b=61, c=u'eight', store=self.store)
        self.one = DataThunk(a=1, b=435716, c=u'one', store=self.store)
        self.two = DataThunk(a=2, b=67145, c=u'two', store=self.store)
        self.four = DataThunk(a=4, b=6327, c=u'four', store=self.store)
        self.five = DataThunk(a=5, b=91856, c=u'five', store=self.store)
        self.scrollFragment = self.getScrollFragment()


    def test_performQueryAscending(self):
        """
        Test that some simple ranges can be correctly retrieved when the sort
        order is ascending on the default column.
        """
        self.scrollFragment.isAscending = True
        for low, high in [(0, 2), (1, 3), (2, 4)]:
            self.assertEquals(
                self.scrollFragment.performQuery(low, high),
                [self.five, self.six, self.seven, self.eight][low:high])


    def test_performQueryDescending(self):
        """
        Like L{test_performQueryAscending} but for the descending sort order.
        """
        self.scrollFragment.isAscending = False
        for low, high in [(0, 2), (1, 3), (2, 4)]:
            self.assertEquals(
                self.scrollFragment.performQuery(low, high),
                [self.eight, self.seven, self.six, self.five][low:high])



class ScrollingFragmentTestCase(ScrollTestMixin,
                                unittest.TestCase):
    """
    Test cases which simulate various client behaviors to exercise the legacy
    L{ScrollingFragment}.
    """
    def getScrollFragment(self):
        sf = ScrollingFragment(
            self.store, DataThunk, DataThunk.a > 4,
            [DataThunk.b, DataThunk.c], DataThunk.a)
        sf.linkToItem = lambda ign: None
        return sf


    def testGetTwoChunks(self):
        self.assertEquals(
            self.scrollFragment.requestRowRange(0, 2),
            [{'c': u'five', 'b': 91856}, {'c': u'six', 'b': 8157}])

        self.assertEquals(
            self.scrollFragment.requestRowRange(2, 4),
            [{'c': u'seven', 'b': 4724}, {'c': u'eight', 'b': 61}])

        self.scrollFragment.resort('b')

        self.assertEquals(self.scrollFragment.requestRowRange(0, 2),
                          [{'c': u'eight', 'b': 61}, {'c': u'seven', 'b': 4724}])
        self.assertEquals(self.scrollFragment.requestRowRange(2, 4),
                          [{'c': u'six', 'b': 8157}, {'c': u'five', 'b': 91856}])


    def testSortsOnFirstSortable(self):
        """
        Test that the scrolltable sorts on the first sortable column
        """
        sf = ScrollingFragment(
                self.store, DataThunk, None,
                (UnsortableColumn(DataThunk.a),
                 DataThunk.b))
        self.assertEquals(sf.currentSortColumn, DataThunk.b)
    testSortsOnFirstSortable.suppress = [_unsortableColumnSuppression]


    def testSortsOnFirstSortable2(self):
        """
        Same as L{testSortsOnFirstSortable}, but for the case where the first
        sortable column is the first in the column list
        """
        sf = ScrollingFragment(
                self.store, DataThunk, None,
                (DataThunk.a, UnsortableColumn(DataThunk.b)))

        self.assertIdentical(sf.currentSortColumn.sortAttribute(),
                             DataThunk.a)
    testSortsOnFirstSortable2.suppress = [_unsortableColumnSuppression]


    def testTestNoSortables(self):
        """
        Test that the scrolltable can handle the case where all columns are
        unsortable
        """
        sf = ScrollingFragment(
                self.store, DataThunk, None,
                (UnsortableColumn(DataThunk.a),
                 UnsortableColumn(DataThunk.b)))

        self.assertEquals(sf.currentSortColumn, None)
    testTestNoSortables.suppress = [_unsortableColumnSuppression]


    def test_unsortableColumnType(self):
        """
        L{UnsortableColumn.getType} should return the same value as
        L{AttributeColumn.getType} for a particular attribute.
        """
        self.assertEqual(
            AttributeColumn(DataThunk.a).getType(),
            UnsortableColumn(DataThunk.a).getType())
    test_unsortableColumnType.suppress = [_unsortableColumnSuppression]


    def test_unsortableColumnDeprecated(self):
        """
        L{UnsortableColumn} is a deprecated almost-alias for
        L{UnsortableColumnWrapper}.
        """
        self.assertWarns(
            DeprecationWarning,
            "Use UnsortableColumnWrapper(AttributeColumn(*a, **kw)) "
            "instead of UnsortableColumn(*a, **kw).",
            __file__,
            lambda: UnsortableColumn(DataThunk.a))


    def testUnsortableColumnWrapper(self):
        """
        Test that an L{UnsortableColumnWrapper} wrapping an L{AttributeColumn}
        is treated the same as L{UnsortableColumn}
        """
        sf = ScrollingFragment(
                self.store, DataThunk, None,
                (UnsortableColumnWrapper(AttributeColumn(DataThunk.a)),
                 DataThunk.b))

        self.assertEquals(sf.currentSortColumn, DataThunk.b)


    def test_allUnsortableSortMetadata(self):
        """
        Test that C{getTableMetadata} is correct with respect to the
        sortability of columns
        """
        sf = ScrollingFragment(
                self.store, DataThunk, None,
                (UnsortableColumn(DataThunk.a),
                 UnsortableColumn(DataThunk.b)))

        meta = sf.getTableMetadata()
        cols = meta[1]
        self.assertEquals(cols['a'][1], False)
        self.assertEquals(cols['b'][1], False)
    test_allUnsortableSortMetadata.suppress = [_unsortableColumnSuppression]


    def test_oneSortableSortMetadata(self):
        """
        Same as L{test_allUnsortableSortMetadata}, but with one sortable column
        """
        sf = ScrollingFragment(
                self.store, DataThunk, None,
                (DataThunk.a,
                 UnsortableColumn(DataThunk.b)))

        meta = sf.getTableMetadata()
        cols = meta[1]
        self.assertEquals(cols['a'][1], True)
        self.assertEquals(cols['b'][1], False)
    test_oneSortableSortMetadata.suppress = [_unsortableColumnSuppression]



class SequenceScrollingFragmentTestCase(ScrollTestMixin, unittest.TestCase):
    """
    Run the general scrolling tests against L{SequenceScrollingFragment}.
    """
    def getScrollFragment(self):
        return SequenceScrollingFragment(
            self.store,
            [self.five, self.six, self.seven, self.eight],
            [DataThunk.b, DataThunk.c], DataThunk.a)


class StoreIDSequenceScrollingFragmentTestCase(ScrollTestMixin, unittest.TestCase):
    """
    Run the general scrolling tests against
    L{StoreIDSequenceScrollingFragmentTestCase}.
    """
    def getScrollFragment(self):
        return StoreIDSequenceScrollingFragment(
            self.store,
            [self.five.storeID, self.six.storeID,
             self.seven.storeID, self.eight.storeID],
            [DataThunk.b, DataThunk.c], DataThunk.a)



class TestableInequalityModel(InequalityModel):
    """
    Helper for InequalityModel tests which implements the row construction
    hook.
    """
    def constructRows(self, items):
        """
        Squash the given items iterator into a list and return it so it can be
        inspected by tests.
        """
        return list(items)



class InequalityModelTestCase(unittest.TestCase):
    """
    Tests for the inequality-based scrolling model implemented by
    L{InequalityModel}.
    """
    def setUp(self):
        """
        Set up an inequality model (by way of L{TestableInequalityModel}) backed by
        a user store with some sample data, and an L{IWebTranslator} powerup,
        to provide a somewhat realistic test setup.

        The data provided has some duplicates in the sort column, and it is
        intentionally inserted out of order, so that storeID order and sort
        column ordering will not coincide.
        """
        self.store = Store()
        privApp = PrivateApplication(store=self.store)
        installOn(privApp, self.store)

        self.data = []
        for a, b, c in [(9, 1928, u'nine'), (1, 983, u'one'),
                        (8, 843, u'eight'), (2, 827, u'two'),
                        # (8, 1874, u'eight (DUP)'), (2, 294, u'two (DUP)'),
                        (7, 18, u'seven'), (3, 19, u'three'),
                        (6, 218, u'six'), (4, 2198, u'four'),
                        (5, 1982, u'five'), (0, 10, u'zero')]:
            self.data.append(DataThunk(store=self.store, a=a, b=b, c=c))
        self.data.sort(key=lambda item: (item.a, item.storeID))

        self.model = TestableInequalityModel(
            self.store,
            DataThunk,
            None,
            [DataThunk.a,
             DataThunk.b,
             DataThunk.c],
            DataThunk.a,
            True)


    def test_noSortableColumns(self):
        """
        Attempting to construct a L{InequalityModel} without any sortable columns
        should result in a L{Unsortable} exception being thrown.
        """
        makeModel = lambda: InequalityModel(self.store, DataThunk, None,
                                            [UnsortableColumn(DataThunk.a)],
                                            None, True)
        self.assertRaises(Unsortable, makeModel)
    test_noSortableColumns = [_unsortableColumnSuppression]


    def test_rowsAtStart(self):
        """
        Verify that retrieving rows after the "None" value will retrieve rows at
        the start of the order.
        """
        count = 5
        expected = self.data[:count]
        self.assertEquals(self.model.rowsAfterValue(None, count),
                          expected)


    def test_rowsAtEnd(self):
        """
        Verify that retrieving rows before the "None" value will retrieve rows
        at the end of the sort order.
        """
        count = 5
        expected = self.data[-count:]
        self.assertEquals(self.model.rowsBeforeValue(None, count),
                          expected)


    def test_rowsAfterValue(self):
        """
        Verify that rows after a particular value with a particular ordering
        are returned by L{InequalityModel.rowsAfterValue}.
        """
        for count in range(1, len(self.data)):
            for value in range(len(self.data)):
                expected = self.data[value:value + count]
                self.assertEqual(
                    self.model.rowsAfterValue(value, count),
                    expected)


    def test_rowsAfterRow(self):
        """
        Verify that the exposed method, L{rowsAfterRow}, returns results comparable
        to the logical method, L{rowsAfterItem}.
        """
        args = []
        def rowsAfterItem(item, count):
            args.append((item, count))
        self.model.rowsAfterItem = rowsAfterItem
        theItem = object()
        theTranslator = FakeTranslator()
        theTranslator.fromWebID = lambda webID: theItem
        self.model.webTranslator = theTranslator
        self.model.rowsAfterRow({u'__id__': u'webid!'}, 10)
        self.assertEqual(args, [(theItem, 10)])


    def test_rowsBeforeRow(self):
        """
        Verify that the exposed method, L{rowsBeforeRow}, return results comparable
        to the logical method, L{rowsBeforeItem}.
        """
        args = []
        def rowsBeforeItem(item, count):
            args.append((item, count))
        self.model.rowsBeforeItem = rowsBeforeItem
        theItem = object()
        theTranslator = FakeTranslator()
        theTranslator.fromWebID = lambda webID: theItem
        self.model.webTranslator = theTranslator
        self.model.rowsBeforeRow({u'__id__': u'webid!'}, 10)
        self.assertEqual(args, [(theItem, 10)])


    def test_rowsAfterItem(self):
        """
        Verify that rows after a particular item with a particular ordering are
        returned by L{InequalityModel.rowsAfterItem}.
        """
        for count in range(1, len(self.data)):
            for value in range(1, len(self.data)):
                expected = self.data[value:value + count]
                self.assertEqual(
                    self.model.rowsAfterItem(self.data[value - 1], count),
                    expected)



    def test_rowsBeforeValue(self):
        """
        Like L{test_rowsAfterValue} but for data going in the opposite
        direction.
        """
        for count in range(1, len(self.data)):
            for value in range(len(self.data)):
                expected = self.data[max(value - count, 0):value]
                self.assertEqual(
                    self.model.rowsBeforeValue(value, count),
                    expected)


    def test_rowsBeforeItem(self):
        """
        Like L{test_rowsAfterItem} but for data going in the opposite
        direction.
        """
        for count in range(1, len(self.data)):
            for value in range(len(self.data)):
                expected = self.data[max(value - count, 0):value]
                self.assertEqual(
                    self.model.rowsBeforeItem(self.data[value], count),
                    expected)



class FakeTranslator:
    """
    A fake implementation of the L{IWebTranslator} interface, for tests which
    do not use this.
    """
    implements(IWebTranslator)
    def fromWebID(self, webID):
        return None
    def toWebID(self, item):
        return "none"
    def linkTo(self, storeID):
        return "none"
    def linkFrom(self, webID):
        return 0


class ScrollingElementTests(unittest.TestCase):
    """
    Test cases for the ultimate client-facing subclass, L{ScrollingElement}.
    """
    def test_deprecatedMissingWebTranslator(self):
        """
        Instantiating a L{ScrollingElement} without supplying an
        L{IWebTranslator}, either explicitly or via a store powerup, will
        emit a L{DeprecationWarning} explaining that this should not be
        done.
        """
        def makeScrollingElement():
            return ScrollingElement(
                Store(), DataThunk, None, [DataThunk.a], DataThunk.a, True)
        self.assertWarns(
            DeprecationWarning,
            "No IWebTranslator plugin when creating Scrolltable - broken "
            "configuration, now deprecated!  Try passing webTranslator "
            "keyword argument.",
            __file__,
            makeScrollingElement)


    def test_callComparableValue(self):
        """
        L{ScrollingElement} should attempt to call L{IColumn.toComparableValue} to
        translate input from JavaScript if it is provided by the sort column.
        """
        calledWithValues = []
        column = AttributeColumn(DataThunk.a)
        column.toComparableValue = lambda v: (calledWithValues.append(v), 0)[1]
        scrollingElement = ScrollingElement(
            Store(), DataThunk, None, [column], webTranslator=FakeTranslator())
        scrollingElement.rowsAfterValue(16, 10)
        self.assertEqual(calledWithValues, [16])
        calledWithValues.pop()
        scrollingElement.rowsBeforeValue(11, 10)
        self.assertEqual(calledWithValues, [11])


    def test_deprecatedNoToComparableValue(self):
        """
        If L{IColumn.toComparableValue} is I{not} provided by the sort column,
        then L{ScrollingElement} should notify the developer of a deprecation
        warning but default to using the value itself.
        """
        class FakeComparator:
            def __ge__(fc, other):
                self.shouldBeFakeValue = other
            __lt__ = __ge__
        theFakeComparator = FakeComparator()
        class FakeOldColumn:
            implements(IColumn) # but not really; we're missing something!
            attributeID = 'fake'
            def sortAttribute(self):
                return theFakeComparator

        scrollingElement = ScrollingElement(
            Store(), DataThunk, None, [FakeOldColumn()],
            webTranslator=FakeTranslator())

       # Now, completely hamstring the implementation; we are interested in
       # something very specific here, so we just want to capture one value.
       # (Passing it further on, for example to Axiom, would not result in
       # testing anything useful, since it is the individual column's
       # responsibility to yield useful values here anyway, and this test is
       # explicitly for the case where it's *not* doing that, but happened to
       # work before anyway!)

        scrollingElement.inequalityQuery = lambda a, b, c: None
        scrollingElement.constructRows = lambda nothing : ()
        for lenientMethod in [scrollingElement.rowsBeforeValue,
                              scrollingElement.rowsAfterValue]:
            fakeValue = object()
            self.assertWarns(DeprecationWarning,
                             "IColumn implementor %s.FakeOldColumn does not "
                             "implement method toComparableValue.  This is "
                             "required since Mantissa 0.6.6." % (__name__,),
                             __file__, lenientMethod, fakeValue, 10)
            self.assertIdentical(fakeValue, self.shouldBeFakeValue)


    def test_initialWidgetArguments(self):
        """
        Verify that the arguments the client widget expects: the name of the
        current sort column, a list of the available data columns, and the sort
        order.
        """
        s = Store()
        testElement = ScrollingElement(s, DataThunk, None,
                                       [DataThunk.a, DataThunk.c],
                                       DataThunk.a, True,
                                       FakeTranslator())
        self.assertEqual(testElement.getInitialArguments(),
                         [u"a",
                         [{u"name": u"a",
                           u"type": u"integer"},
                          {u"name": u"c",
                           u"type": u"text"}],
                          True])


    def test_missingTypeDefaultsToText(self):
        """
        When constructed with an L{IColumn} which returns C{None} from its
        C{getType} method, L{ScrollingElement} should use the default of
        C{text} for that column's type.
        """
        class UntypedColumn(object):
            implements(IColumn)
            attributeID = 'foo'
            def sortAttribute(self):
                return None
            def getType(self):
                return None
        column = UntypedColumn()

        scroller = ScrollingElement(
            None, None, None, [DataThunk.a, column], None,
            webTranslator=object())
        attribute, columnList, ascending = scroller.getInitialArguments()
        self.assertEqual(
            columnList,
            [{u'type': u'integer', u'name': u'a'},
             {u'type': u'text', u'name': u'foo'}])



class InequalityModelDuplicatesTestCase(unittest.TestCase):
    """
    Similar to L{InequalityModelTestCase}, but test cases where there are
    multiple rows with the same value for the sort key.
    """
    def setUp(self):
        self.store = Store()
        privApp = PrivateApplication(store=self.store)
        installOn(privApp, self.store)

        # Create some data to test with in an order which is not naturally
        # sorted in any way.

        # 4 1s, 4 0s, 4 2s
        self.data = []
        for number in [1, 0, 2, 2, 0, 1, 1, 0, 2, 0, 2, 1]:
            self.data.append(DataThunk(store=self.store, a=number))

        # But order it for the sake of simplicity while testing.
        self.data.sort(key=lambda item: (item.a, item.storeID))

        self.model = TestableInequalityModel(
            self.store,
            DataThunk,
            None,
            [DataThunk.a,
             DataThunk.b,
             DataThunk.c],
            DataThunk.a,
            True)


    def test_rowsAfterValue(self):
        """
        Test that when there are duplicate values in the sort column,
        L{InequalityModel.rowsAfterValue} returns all the rows (up to the
        indicated maximum) with the value it is passed and then any rows with
        values greater than the given value.
        """
        self.assertEqual(
            self.model.rowsAfterValue(1, 3),
            self.data[4:4+3])


    def test_rowsAfterItem(self):
        """
        Like L{test_rowsAfterValue}, but for L{InequalityModel.rowsAfterItem}
        method.
        """
        # Test that starting at the first item with a particular value and
        # crossing over to another value works properly.
        self.assertEqual(
            self.model.rowsAfterItem(self.data[0], 3),
            self.data[1:4])
        # Test that starting at an item "in the middle" of a particular value
        # and crossing over to another value works properly.
        self.assertEqual(
            self.model.rowsAfterItem(self.data[1], 3),
            self.data[2:5])
        # Test that starting at the first item with a particular value and not
        # requesting enough rows to cross into another value works properly.
        self.assertEqual(
            self.model.rowsAfterItem(self.data[0], 1),
            [self.data[1]])


    def test_rowsBeforeValue(self):
        """
        Like L{test_rowsAfterValue}, but for L{InequalityModel.rowsBeforeValue}
        """
        self.assertEqual(
            self.model.rowsBeforeValue(2, 3),
            self.data[5:8])


    def test_rowsBeforeItem(self):
        """
        Like L{test_rowsAfterItem}, but for L{InequalityModel.rowsBeforeItem}.
        """
        for x in range(len(self.data)):
            self.assertEqual(
                self.model.rowsBeforeItem(self.data[x], 20),
                self.data[:x])




class InequalityPerformanceTests(unittest.TestCase):
    """
    Tests for the complexity and runtime costs of the methods of
    L{InequalityModel}.
    """
    def setUp(self):
        self.store = Store()
        privApp = PrivateApplication(store=self.store)
        installOn(privApp, self.store)
        self.model = TestableInequalityModel(
            self.store,
            DataThunkWithIndex,
            None,
            [DataThunkWithIndex.a],
            DataThunkWithIndex.a,
            True,
            privApp)
        self.counter = QueryCounter(self.store)
        self.data = []
        for i in range(4):
            self.data.append(DataThunkWithIndex(store=self.store, a=i))


    def rowsAfterValue(self, value, count):
        return self.counter.measure(self.model.rowsAfterValue, value, count)


    def rowsAfterItem(self, item, count):
        return self.counter.measure(self.model.rowsAfterItem, item, count)


    def rowsBeforeValue(self, value, count):
        return self.counter.measure(self.model.rowsBeforeValue, value, count)


    def rowsBeforeItem(self, item, count):
        return self.counter.measure(self.model.rowsBeforeItem, item, count)


    def test_rowsAfterValue(self):
        """
        Verify that the cost of L{InequalityModel.rowsAfterValue} is
        independent of the total number of rows in the table being queried, as
        long as that number is greater than the number of rows requested.
        """
        first = self.rowsAfterValue(1, 2)
        DataThunkWithIndex(store=self.store, a=4)
        second = self.rowsAfterValue(1, 2)
        self.assertEqual(first, second)


    def test_rowsAfterValueWithDuplicatesBeforeStart(self):
        """
        Like L{test_rowsAfterValue}, but verify the behavior in the face of
        duplicate rows before the start value.
        """
        first = self.rowsAfterValue(1, 2)
        DataThunkWithIndex(store=self.store, a=0)
        second = self.rowsAfterValue(1, 2)
        self.assertEqual(first, second)


    def test_rowsAfterValueWithDuplicatesAtStart(self):
        """
        Like L{test_rowsAfterValue}, but verify the behavior in the face of
        duplicate rows exactly at the start value.
        """
        first = self.rowsAfterValue(1, 2)
        DataThunkWithIndex(store=self.store, a=1)
        second = self.rowsAfterValue(1, 2)
        self.assertEqual(first, second)


    def test_rowsAfterValueWithDuplicatesInResult(self):
        """
        Like L{test_rowsAfterValue}, but verify the behavior in the face of
        duplicate rows in the result set.
        """
        first = self.rowsAfterValue(1, 2)
        DataThunkWithIndex(store=self.store, a=2)
        second = self.rowsAfterValue(1, 2)
        self.assertEqual(first, second)


    def test_rowsAfterValueWithDuplicatesAfter(self):
        """
        Like L{test_rowsAfterValue}, but verify the behavior in the face of
        duplicate rows past the end of the result set.
        """
        first = self.rowsAfterValue(1, 2)
        DataThunkWithIndex(store=self.store, a=4)
        second = self.rowsAfterValue(1, 2)
        self.assertEqual(first, second)


    def test_rowsAfterItem(self):
        """
        Like L{test_rowsAfterValue}, but for L{InequalityModel.rowsAfterItem}.
        """
        first = self.rowsAfterItem(self.data[0], 2)
        DataThunkWithIndex(store=self.store, a=4)
        second = self.rowsAfterItem(self.data[0], 2)
        self.assertEqual(first, second)


    def test_rowsAfterItemWithDuplicatesBeforeStart(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesBeforeStart}, but for
        L{InequalityModel.rowsAfterItem}.
        """
        DataThunkWithIndex(store=self.store, a=-1)
        first = self.rowsAfterItem(self.data[0], 2)
        DataThunkWithIndex(store=self.store, a=-1)
        second = self.rowsAfterItem(self.data[0], 2)
        self.assertEqual(first, second)


    def test_rowsAfterItemWithDuplicatesAtStart(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesAtStart}, but for
        L{InequalityModel.rowsAfterItem}.
        """
        first = self.rowsAfterItem(self.data[0], 2)
        DataThunkWithIndex(store=self.store, a=0)
        second = self.rowsAfterItem(self.data[0], 2)
        self.assertEqual(first, second)
    test_rowsAfterItemWithDuplicatesAtStart.todo = (
        "Index scan to find appropriate storeID starting point once the "
        "value index has been used to seek to /near/ the correct starting "
        "place causes this to be O(N) on the number of rows with duplicate "
        "values.")


    def test_rowsAfterItemWithDuplicatesInResult(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesInResult}, but for
        L{InequalityModel.rowsAfterItem}.
        """
        first = self.rowsAfterItem(self.data[0], 2)
        DataThunkWithIndex(store=self.store, a=1)
        second = self.rowsAfterItem(self.data[0], 2)
        self.assertEqual(first, second)


    def test_rowsAfterItemWithDuplicatesAfter(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesAfter}, but for
        L{InequalityModel.rowsAfterItem}.
        """
        first = self.rowsAfterItem(self.data[0], 2)
        DataThunkWithIndex(store=self.store, a=3)
        second = self.rowsAfterItem(self.data[0], 2)
        self.assertEqual(first, second)


    def test_rowsBeforeValue(self):
        """
        Like L{test_rowsAfterValue}, but for
        L{InequalityModel.rowsBeforeValue}.
        """
        first = self.rowsBeforeValue(2, 2)
        DataThunkWithIndex(store=self.store, a=-1)
        second = self.rowsBeforeValue(2, 2)
        self.assertEqual(first, second)


    def test_rowsBeforeValueWithDuplicatesBeforeStart(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesBeforeStart}, but for
        L{InequalityModel.rowsBeforeValue}.
        """
        first = self.rowsBeforeValue(2, 2)
        DataThunkWithIndex(store=self.store, a=3)
        second = self.rowsBeforeValue(2, 2)
        self.assertEqual(first, second)


    def test_rowsBeforeValueWithDuplicatesAtStart(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesAtStart}, but for
        L{InequalityModel.rowsBeforeValue}.
        """
        first = self.rowsBeforeValue(2, 2)
        DataThunkWithIndex(store=self.store, a=2)
        second = self.rowsBeforeValue(2, 2)
        self.assertEqual(first, second)


    def test_rowsBeforeValueWithDuplicatesInResult(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesInResult}, but for
        L{InequalityModel.rowsBeforeValue}.
        """
        first = self.rowsBeforeValue(2, 2)
        DataThunkWithIndex(store=self.store, a=1)
        second = self.rowsBeforeValue(2, 2)
        self.assertEqual(first, second)


    def test_rowsBeforeValueWithDuplicatesAfter(self):
        """
        Like L{test_rowsAfterValueWithDuplicatesAfter}, but for
        L{InequalityModel.rowsBeforeValue}.
        """
        first = self.rowsBeforeValue(2, 2)
        DataThunkWithIndex(store=self.store, a=0)
        second = self.rowsBeforeValue(2, 2)
        self.assertEqual(first, second)


    def test_rowsBeforeItem(self):
        """
        Like L{test_rowsAfterItem}, but for L{InequalityModel.rowsBeforeItem}.
        """
        first = self.rowsBeforeItem(self.data[3], 2)
        DataThunkWithIndex(store=self.store, a=-1)
        second = self.rowsBeforeItem(self.data[3], 2)
        self.assertEqual(first, second)


    def test_rowsBeforeItemWithDuplicatesBeforeStart(self):
        """
        Like L{test_rowsAfterItemWithDuplicatesBeforeStart}, but for
        L{InequalityModel.rowsBeforeItem}.
        """
        DataThunkWithIndex(store=self.store, a=4)
        first = self.rowsBeforeItem(self.data[3], 2)
        DataThunkWithIndex(store=self.store, a=4)
        second = self.rowsBeforeItem(self.data[3], 2)
        self.assertEqual(first, second)


    def test_rowsBeforeItemWithDuplicatesAtStart(self):
        """
        Like L{test_rowsAfterItemWithDuplicatesAtStart}, but for
        L{Inequality.rowsBeforeItem}.
        """
        first = self.rowsBeforeItem(self.data[3], 2)
        DataThunkWithIndex(store=self.store, a=3)
        second = self.rowsBeforeItem(self.data[3], 2)
        self.assertEqual(first, second)
    test_rowsBeforeItemWithDuplicatesAtStart.todo = (
        "Index scan to find appropriate storeID starting point once the "
        "value index has been used to seek to /near/ the correct starting "
        "place causes this to be O(N) on the number of rows with duplicate "
        "values.")


    def test_rowsBeforeItemWithDuplicatesInResult(self):
        """
        Like L{test_rowsAfterItemWithDuplicatesInResult}, but for
        L{Inequality.rowsBeforeItem}.
        """
        first = self.rowsBeforeItem(self.data[3], 2)
        DataThunkWithIndex(store=self.store, a=2)
        second = self.rowsBeforeItem(self.data[3], 2)
        self.assertEqual(first, second)


    def test_rowsBeforeItemWithDuplicatesAfter(self):
        """
        Like L{test_rowsAfterItemWithDuplicatesAfter}, but for
        L{InequalityModel.rowsBeforeItem}.
        """
        first = self.rowsBeforeItem(self.data[3], 2)
        DataThunkWithIndex(store=self.store, a=0)
        second = self.rowsBeforeItem(self.data[3], 2)
        self.assertEqual(first, second)




class UnsortableColumnWrapperTestCase(unittest.TestCase):
    """
    Tests for L{UnsortableColumnWrapper}
    """

    def test_unsortableColumnWrapper(self):
        attr = DataThunk.a
        col = AttributeColumn(attr)
        unsortableCol = UnsortableColumnWrapper(col)

        item = DataThunk(store=Store(), a=26)

        value = unsortableCol.extractValue(None, item)
        self.assertEquals(value, item.a)
        self.assertEquals(value, col.extractValue(None, item))

        typ = unsortableCol.getType()
        self.assertEquals(typ, 'integer')
        self.assertEquals(typ, col.getType())

        self.assertEquals(unsortableCol.sortAttribute(), None)



# Copyright 2008 Divmod, Inc. See LICENSE file for details
# -*- test-case-name: xmantissa.test.test_webapp,xmantissa.test.test_publicweb,xmantissa.test.test_website -*-

"""
This unfortunate module exists to contain code that would create an ugly
dependency loop if it were somewhere else.
"""
from zope.interface import implements

from twisted.cred.portal import IRealm

from epsilon.structlike import record

from axiom.userbase import getDomainNames

from nevow import athena
from nevow.rend import NotFound
from nevow.inevow import IResource, IRequest

from xmantissa.ixmantissa import (IWebViewer, INavigableFragment,
                                  ISiteRootPlugin)

from xmantissa.websharing import UserIndexPage

from xmantissa.error import CouldNotLoadFromThemes


class WebViewerHelper(object):
    """
    This is a mixin for the common logic in the two providers of
    L{IWebViewer} included with Mantissa,
    L{xmantissa.publicweb._AnonymousWebViewer} and
    L{xmantissa.webapp._AuthenticatedWebViewer}.

    @ivar _getDocFactory: a 1-arg callable which returns a nevow loader.

    @ivar _preferredThemes: a 0-arg callable which returns a list of nevow
        themes.
    """
    def __init__(self, _getDocFactory, _preferredThemes):
        """
        """
        self._getDocFactory = _getDocFactory
        self._preferredThemes = _preferredThemes


    def _wrapNavFrag(self, fragment, useAthena):
        """
        Subclasses must implement this to wrap a fragment.

        @param fragment: an L{INavigableFragment} provider that should be
            wrapped in the resulting page.

        @param useAthena: Whether the resulting L{IResource} should be a
            L{LivePage}.

        @type useAthena: L{bool}

        @return: a fragment to display to the user.

        @rtype: L{IResource}
        """


    def wrapModel(self, model):
        """
        Converts application-provided model objects to L{IResource} providers.
        """
        res = IResource(model, None)
        if res is None:
            frag = INavigableFragment(model)
            fragmentName = getattr(frag, 'fragmentName', None)
            if fragmentName is not None:
                fragDocFactory = self._getDocFactory(fragmentName)
                if fragDocFactory is not None:
                    frag.docFactory = fragDocFactory
            if frag.docFactory is None:
                raise CouldNotLoadFromThemes(frag, self._preferredThemes())
            useAthena = isinstance(frag, (athena.LiveFragment, athena.LiveElement))
            return self._wrapNavFrag(frag, useAthena)
        else:
            return res



class MantissaViewHelper(object):
    """
    This is the superclass of all Mantissa resources which act as a wrapper
    around an L{INavigableFragment} provider.  This must be mixed in to some
    hierarchy with a C{locateChild} method, since it expects to cooperate in
    such a hierarchy.

    Due to infelicities in the implementation of some (pre-existing)
    subclasses, there is no __init__; but subclasses must set the 'fragment'
    attribute in theirs.
    """
    fragment = None

    def locateChild(self, ctx, segments):
        """
        Attempt to locate the child via the '.fragment' attribute, then fall
        back to normal locateChild behavior.
        """
        if self.fragment is not None:
            # There are still a bunch of bogus subclasses of this class, which
            # are used in a variety of distasteful ways.  'fragment' *should*
            # always be set to something that isn't None, but there's no way to
            # make sure that it will be for the moment.  Every effort should be
            # made to reduce public use of subclasses of this class (instead
            # preferring to wrap content objects with
            # IWebViewer.wrapModel()), so that the above check can be
            # removed. -glyph
            lc = getattr(self.fragment, 'locateChild', None)
            if lc is not None:
                x = lc(ctx, segments)
                if x is not NotFound:
                    return x
        return super(MantissaViewHelper, self).locateChild(ctx, segments)



class SiteRootMixin(object):
    """
    Common functionality for L{AnonymousSite} and L{WebSite}.
    """

    def locateChild(self, context, segments):
        """
        Return a statically defined child or a child defined by a site root
        plugin or an avatar from guard.
        """
        request = IRequest(context)
        webViewer = IWebViewer(self.store, None)
        childAndSegments = self.siteProduceResource(request, segments, webViewer)
        if childAndSegments is not None:
            return childAndSegments
        return NotFound


    # IMantissaSite
    def siteProduceResource(self, req, segments, webViewer):
        """
        Retrieve a child resource and segments from rootChild_ methods on this
        object and SiteRootPlugins.

        @return: a 2-tuple of (resource, segments), suitable for return from
        locateChild.

        @param req: an L{IRequest} provider.

        @param segments: a tuple of L{str}s, the segments from the request.

        @param webViewer: an L{IWebViewer}, to be propagated through the child
        lookup process.
        """

        # rootChild_* is not the same as child_, because its signature is
        # different.  Maybe this should be done some other way.
        shortcut = getattr(self, 'rootChild_' + segments[0], None)
        if shortcut:
            res = shortcut(req, webViewer)
            if res is not None:
                return res, segments[1:]

        for plg in self.store.powerupsFor(ISiteRootPlugin):
            produceResource = getattr(plg, 'produceResource', None)
            if produceResource is not None:
                childAndSegments = produceResource(req, segments, webViewer)
            else:
                childAndSegments = plg.resourceFactory(segments)
            if childAndSegments is not None:
                return childAndSegments
        return None


    # IPowerupIndirector
    def indirect(self, interface):
        """
        Create a L{VirtualHostWrapper} so it can have the first chance to
        handle web requests.
        """
        if interface is IResource:
            siteStore = self.store.parent
            if self.store.parent is None:
                siteStore = self.store
            return VirtualHostWrapper(
                siteStore,
                IWebViewer(self.store),
                self)
        return self



class VirtualHostWrapper(record('siteStore webViewer wrapped')):
    """
    Resource wrapper which implements per-user virtual subdomains.  This should
    be wrapped around any resource which sits at the root of the hierarchy.  It
    will examine requests for their hostname and, when appropriate, redirect
    handling of the query to the appropriate sharing resource.

    @type siteStore: L{Store}
    @ivar siteStore: The site store which will be queried to determine which
        hostnames are associated with this server.

    @type webViewer: L{IWebViewer}
    @ivar webViewer: The web viewer representing the user.

    @type wrapped: L{IResource} provider
    @ivar wrapped: A resource to which traversal will be delegated if the
        request is not for a user subdomain.
    """
    implements(IResource)

    def subdomain(self, hostname):
        """
        Determine of which known domain the given hostname is a subdomain.

        @return: A two-tuple giving the subdomain part and the domain part or
            C{None} if the domain is not a subdomain of any known domain.
        """
        hostname = hostname.split(":")[0]
        for domain in getDomainNames(self.siteStore):
            if hostname.endswith("." + domain):
                username = hostname[:-len(domain) - 1]
                if username != "www":
                    return username, domain
        return None


    def locateChild(self, context, segments):
        """
        Delegate dispatch to a sharing resource if the request is for a user
        subdomain, otherwise fall back to the wrapped resource's C{locateChild}
        implementation.
        """
        request = IRequest(context)
        hostname = request.getHeader('host')

        info = self.subdomain(hostname)
        if info is not None:
            username, domain = info
            index = UserIndexPage(IRealm(self.siteStore),
                                  self.webViewer)
            resource = index.locateChild(None, [username])[0]
            return resource, segments
        return self.wrapped.locateChild(context, segments)




"""
Tests for xmantissa._recordattr module.
"""

from twisted.trial.unittest import TestCase

from epsilon.structlike import record

from axiom.item import Item

from axiom.attributes import text, integer

from axiom.store import Store

from xmantissa._recordattr import RecordAttribute, WithRecordAttributes

class Sigma(record('left right')):
    """
    A simple record type which composes two attributes.
    """
    def getLeft(self):
        """
        Return the left attribute.
        """
        return self.left

    def getRight(self):
        """
        Return the right attribute.
        """
        return self.right


class RecordAttributeTestItem(Item):
    """
    An item for testing record attributes.
    """

    alpha = text()
    beta = integer()
    sigma = RecordAttribute(Sigma, [alpha, beta])


class RecordAttributeRequired(Item, WithRecordAttributes):
    """
    An item for testing record attributes with mandatory attributes.
    """
    alpha = text(allowNone=False)
    beta = integer(allowNone=False)
    sigma = RecordAttribute(Sigma, [alpha, beta])



class ItemWithRecordAttributeTest(TestCase):
    """
    An item with a RecordAttribute attribute ought to return and store its
    attributes.
    """

    def eitherWay(self, thunk, T, **kw):
        """
        Run the given 'thunk' with an item not inserted into a store, then one
        inserted into a store after the fact, then one inserted into a store in
        its constructor.
        """
        ra1 = T(**kw)
        thunk(ra1)
        s = Store()
        ra2 = T(**kw)
        ra2.store = s
        thunk(ra2)
        ra3 = T(store=s, **kw)
        thunk(ra3)


    def test_getAttribute(self):
        """
        Retrieving a L{RecordAttribute} whose component attributes are set
        normally should yield a record instance of the appropriate type with
        the values set.
        """
        def check(rati):
            self.assertEqual(rati.sigma.getLeft(), u'one')
            self.assertEqual(rati.sigma.getRight(), 2)
        self.eitherWay(check, RecordAttributeTestItem, alpha=u'one', beta=2)


    def test_initialize(self):
        """
        Initializing an item with a record attribute set to something should
        set its underlying attributes.
        """
        def check(rati):
            self.assertEqual(rati.alpha, u'five')
            self.assertEqual(rati.beta, 6)
        self.eitherWay(check, RecordAttributeTestItem,
                       sigma=Sigma(left=u'five', right=6))


    def test_initializeNoNones(self):
        """
        Initializing an item with a record attribute set to something should
        set its underlying attributes in the case where the underlying item
        does not work.
        """
        def check(rati):
            self.assertEqual(rati.alpha, u'seven')
            self.assertEqual(rati.beta, 8)
        self.eitherWay(check, RecordAttributeRequired.create,
                       sigma=Sigma(left=u'seven', right=8))


    def test_setAttribute(self):
        """
        Setting a L{RecordAttribute} should set all of its component
        attributes.
        """
        def check(rati):
            rati.sigma = Sigma(left=u'three', right=4)
            self.assertEqual(rati.alpha, u'three')
            self.assertEqual(rati.beta, 4)
        self.eitherWay(check, RecordAttributeTestItem)


    def test_queryComparisons(self):
        """
        Querying with an inequality on a L{RecordAttribute} should yield the
        same results as querying on its AND'ed component attributes.
        """
        s = Store()

        RARc = RecordAttributeRequired.create
        x = RARc(store=s, sigma=Sigma(left=u'x', right=1))
        y = RARc(store=s, sigma=Sigma(left=u'y', right=2))
        z = RARc(store=s, sigma=Sigma(left=u'z', right=3))
        a = RARc(store=s, sigma=Sigma(left=u'a', right=4))

        self.assertEqual(list(s.query(
                    RecordAttributeRequired,
                    RecordAttributeRequired.sigma == Sigma(u'z', 3))),
                         [z])
        self.assertEqual(list(s.query(
                    RecordAttributeRequired,
                    RecordAttributeRequired.sigma == Sigma(u'z', 9))),
                         [])
        self.assertEqual(list(s.query(
                    RecordAttributeRequired,
                    RecordAttributeRequired.sigma != Sigma(u'y', 2),
                    sort=RecordAttributeRequired.storeID.ascending)),
                         [x, z, a])







from axiom.userbase import LoginSystem
from axiom.test.historic.stubloader import StubbedTest

from nevow.inevow import IResource

from xmantissa.ixmantissa import IMantissaSite, IWebViewer

from xmantissa.publicweb import AnonymousSite

class AnonymousSiteUpgradeTests(StubbedTest):
    """
    Tests to verify that the L{AnonymousSite} was properly upgraded.
    """

    def test_attribute(self):
        """
        Make sure that the one attribute defined by L{AnonymousSite} is
        properly set.
        """
        ls = self.store.findUnique(LoginSystem)
        site = self.store.findUnique(AnonymousSite)
        self.assertIdentical(site.loginSystem, ls)


    def test_powerups(self):
        """
        L{AnonymousSite} should be installed as a powerup for
        L{IWebViewer} and L{IMantissaSite}.
        """
        self.assertEqual(
            set(list(self.store.interfacesFor(
                        self.store.findUnique(AnonymousSite)))),
            set([IResource, IMantissaSite, IWebViewer]))




"""
Tests for upgrading a WebSite to move the L{IProtocolFactoryFactory} parts onto
a separate item.
"""

from xmantissa.test.historic.test_website4to5 import WebSiteUpgradeTests

# Subclass it to make a TestCase with a __module__ which won't confuse trial
# and to make stub discovery work correctly.  These two things are implicitly
# discovered from the class definition are unfortunate. -exarkun
class WebSiteUpgradeTests6(WebSiteUpgradeTests):
    pass



from itertools import chain
from twisted.python.reflect import namedAny, qual
from twisted.python.components import registerAdapter

from axiom.item import Item
from axiom.attributes import textlist, integer, boolean
from axiom.dependency import installOn, uninstallFrom, installedRequirements

from nevow import athena, tags
from nevow.page import renderer

from xmantissa.suspension import  unsuspendTabProviders
from xmantissa.ixmantissa import INavigableElement, INavigableFragment
from xmantissa import liveform
from xmantissa.offering import getInstalledOfferings
from xmantissa.webnav import Tab

from zope.interface import implements

class Product(Item):
    """
    I represent a collection of powerups to install on a user store. When a
    user is to be endowed with the functionality described here, an
    Installation is created in its store based on me.
    """

    types = textlist()

    def installProductOn(self, userstore):
        """
        Creates an Installation in this user store for our collection
        of powerups, and then install those powerups on the user's
        store.
        """

        def install():
            i = Installation(store=userstore)
            i.types = self.types
            i.install()
        userstore.transact(install)

    def removeProductFrom(self, userstore):
        """
        Uninstall all the powerups this product references and remove
        the Installation item from the user's store. Doesn't remove
        the actual powerups currently, but /should/ reactivate them if
        this product is reinstalled.
        """
        def uninstall():
            #this is probably highly insufficient, but i don't know the
            #requirements
            i = userstore.findFirst(Installation,
                                    Installation.types == self.types)
            i.uninstall()
            i.deleteFromStore()
        userstore.transact(uninstall)

    def installOrResume(self, userstore):
        """
        Install this product on a user store. If this product has been
        installed on the user store already and the installation is suspended,
        it will be resumed. If it exists and is not suspended, an error will be
        raised.
        """
        for i in userstore.query(Installation, Installation.types == self.types):
            if i.suspended:
                unsuspendTabProviders(i)
                return
            else:
                raise RuntimeError("installOrResume called for an"
                                   " installation that isn't suspended")
        else:
            self.installProductOn(userstore)

class Installation(Item):
    """
    I represent a collection of functionality installed on a user store. I
    reference a collection of powerups, probably designated by a Product.
    """

    types = textlist()
    _items = textlist()
    suspended = boolean(default=False)

    def items(self):
        """
        Loads the items this Installation refers to.
        """
        for id in self._items:
            yield self.store.getItemByID(int(id))
    items = property(items)
    def allPowerups(self):
        return set(chain(self.items, *[installedRequirements(self.store, i) for
                                       i in self.items]))
    allPowerups = property(allPowerups)

    def install(self):
        """
        Called when installed on the user store. Installs my powerups.
        """
        items = []
        for typeName in self.types:
            it = self.store.findOrCreate(namedAny(typeName))
            installOn(it, self.store)
            items.append(str(it.storeID).decode('ascii'))
        self._items = items

    def uninstall(self):
        """
        Called when uninstalled from the user store. Uninstalls all my
        powerups.
        """
        for item in self.items:
            uninstallFrom(item, self.store)
        self._items = []


class ProductConfiguration(Item):
    implements(INavigableElement)
    attribute = integer(doc="It is an attribute")

    powerupInterfaces = (INavigableElement,)

    def getTabs(self):
        return [Tab('Admin', self.storeID, 0.5,
                    [Tab('Products', self.storeID, 0.6)],
                    authoritative=False)]

    def createProduct(self, powerups):
        """
        Create a new L{Product} instance which confers the given
        powerups.

        @type powerups: C{list} of powerup item types

        @rtype: L{Product}
        @return: The new product instance.
        """
        types = [qual(powerup).decode('ascii')
                       for powerup in powerups]
        for p in self.store.parent.query(Product):
            for t in types:
                if t in p.types:
                    raise ValueError("%s is already included in a Product" % (t,))
        return Product(store=self.store.parent,
                       types=types)

class ProductFragment(athena.LiveElement):
    fragmentName = 'product-configuration'
    live = 'athena'

    def __init__(self, configger):
        athena.LiveElement.__init__(self)
        self.original = configger

    def head(self):
        #XXX put this in its own CSS file?
        return tags.style(type='text/css')['''
        input[name=linktext], input[name=subject], textarea[name=blurb] { width: 40em }
        ''']

    def getInstallablePowerups(self):
                for installedOffering in getInstalledOfferings(self.original.store.parent).itervalues():
                    for p in installedOffering.installablePowerups:
                        yield p


    def coerceProduct(self, **kw):
        """
        Create a product and return a status string which should be part of a
        template.

        @param **kw: Fully qualified Python names for powerup types to
        associate with the created product.
        """
        self.original.createProduct(filter(None, kw.values()))
        return u'Created.'


    def makePowerupCoercer(self, powerup):
        def powerupCoercer(selectedPowerup):
            if selectedPowerup:
                return powerup
            else:
                return None
        return powerupCoercer

    def makePowerupSelector(self, desc):
        return liveform.Parameter('selectedPowerup',
                                  liveform.CHECKBOX_INPUT,
                                  bool, desc)

    def powerupConfigurationParameter(self, (name, desc, p)):
        return liveform.Parameter(
            name,
            liveform.FORM_INPUT,
            liveform.LiveForm(self.makePowerupCoercer(p),
                              [self.makePowerupSelector(desc)],
                              name))
    def productConfigurationForm(self, request, tag):
        productList = liveform.LiveForm(self.coerceProduct,
                                        [self.powerupConfigurationParameter(pi)
                                         for pi in self.getInstallablePowerups()],
                                        u"Installable Powerups")
        productList.setFragmentParent(self)
        return productList

    renderer(productConfigurationForm)

    def configuredProducts(self, request, tag):
        for prod in self.original.store.parent.query(Product):
            yield repr(prod.types)

    renderer(configuredProducts)
registerAdapter(ProductFragment, ProductConfiguration, INavigableFragment)



# -*- test-case-name: xmantissa.test.test_scrolltable -*-

"""
Scrollable tabular data-display area.

This module provides an API for displaying data from a Twisted server in a
Nevow Athena front-end.
"""

import inspect
import warnings

from zope.interface import implements

from twisted.python.components import registerAdapter
from twisted.python.reflect import qual

from epsilon.extime import Time

from nevow.athena import LiveElement, expose

from axiom.attributes import timestamp, SQLAttribute, AND

from xmantissa.ixmantissa import IWebTranslator, IColumn
from xmantissa.error import Unsortable



TYPE_FRAGMENT = 'fragment'
TYPE_WIDGET = 'widget'



class AttributeColumn(object):
    """
    Implement a mapping between Axiom attributes and the scrolltable-based
    L{IColumn}.
    """
    implements(IColumn)

    def __init__(self, attribute, attributeID=None):
        """
        Create an L{AttributeColumn} from an Axiom attribute.

        @param attribute: an axiom L{SQLAttribute} subclass.

        @param attributeID: an optional client-side identifier for this
        attribute.  Normally this will be this attribute's name; it isn't
        visible to the user on the client, it's simply the client-side internal
        identifier.
        """
        self.attribute = attribute
        if attributeID is None:
            attributeID = attribute.attrname
        self.attributeID = attributeID


    def extractValue(self, model, item):
        """
        Extract a simple value for this column from a given item, suitable for
        serialization via Athena's client-communication layer.

        @param model: The scrollable view object requesting the value.
        Unfortunately due to the long history of this code, this has no clear
        interface, and might be a L{ScrollingElement}, L{ScrollingFragment}, or
        L{xmantissa.tdb.TabularDataModel}, depending on which type this
        L{AttributeColumn} was passed to.

        @param item: An instance of the class that this L{AttributeColumn}'s
        L{attribute} was taken from, to retrieve the value from.

        @return: a value of an attribute of C{item}, of a type dependent upon
        this L{AttributeColumn}'s L{attribute}.
        """
        return self.attribute.__get__(item)


    def sortAttribute(self):
        """
        @return: an L{axiom.attributes.Comparable} that can be used to adjust an
        axiom query to sort the table by this column, or None, if this column
        cannot be sorted by.
        """
        return self.attribute


    def getType(self):
        """
        @return: a string to identify the browser-side type of this column to the
        JavaScript code in Mantissa.ScrollTable.ScrollTable.__init__.
        @rtype: L{str}
        """
        sortattr = self.sortAttribute()
        if sortattr is not None:
            return sortattr.__class__.__name__


    def toComparableValue(self, value):
        """
        Convert C{value} into something that can be compared like-for-like with
        L{sortAttribute}.
        """
        return value



registerAdapter(AttributeColumn, SQLAttribute, IColumn)



# these objects aren't for view junk - they allow the model
# to inform the javascript controller about which columns are
# sortable, as well as supporting non-attribute columns

class TimestampAttributeColumn(AttributeColumn):
    """
    Timestamps are a special case; we need to get the posix timestamp so we can
    send the attribute value to javascript.  we don't register an adapter for
    attributes.timestamp because the TDB model uses IColumn.extractValue() to
    determine the value of the query pivot, and so it needs an extime.Time
    instance, not a float.
    """
    def extractValue(self, model, item):
        val = AttributeColumn.extractValue(self, model, item)
        if val is None:
            raise AttributeError("%r was None" % (self.attribute,))
        return val.asPOSIXTimestamp()


    def getType(self):
        return 'timestamp'


    def toComparableValue(self, value):
        """
        Override L{AttributeColumn}'s implementation to return a L{Time} instance.
        """
        return Time.fromPOSIXTimestamp(value)



class UnsortableColumnWrapper(object):
    """
    Wraps an L{AttributeColumn} and makes it unsortable

    @ivar column: L{AttributeColumn}
    """
    implements(IColumn)

    def __init__(self, column):
        self.column = column
        self.attribute = column.attribute
        self.attributeID = column.attributeID


    def extractValue(self, model, item):
        """
        Delegate to the wrapped column's value extraction method.
        """
        return self.column.extractValue(model, item)


    def sortAttribute(self):
        """
        Prevent sorting on this column by ignoring the wrapped column's sort
        attribute and always returning C{None}.
        """
        return None


    def getType(self):
        """
        Return the wrapped column's type.
        """
        return self.column.getType()



class UnsortableColumn(AttributeColumn):
    """
    An axiom attribute column which does not allow server-side sorting for
    policy or performance reasons.
    """
    def __init__(self, *a, **kw):
        warnings.warn(
            category=DeprecationWarning,
            message=(
                "Use UnsortableColumnWrapper(AttributeColumn(*a, **kw)) instead "
                "of UnsortableColumn(*a, **kw)."),
            stacklevel=2)
        AttributeColumn.__init__(self, *a, **kw)


    def sortAttribute(pelf):
        """
        UnsortableColumns are not sortable, so this will always return L{None}.
        See L{AttributeColumn.sortAttribute}.
        """
        return None


    def getType(self):
        """
        Clobber the inherited implementation to work around the fact that
        sortAttribute returns None so that a useful value is still returned.
        """
        return self.attribute.__class__.__name__



class _ScrollableBase(object):
    """
    _ScrollableBase is an internal base class holding logic for dealing with
    lists of L{xmantissa.ixmantissa.IColumn}s, sorting, and link generation.

    This logic is shared by two quite different implementations of
    client-server communication about rows: L{InequalityModel}, which uses
    techniques specific to the performance characteristics of Axiom queries,
    and L{IndexingModel}, which uses simple indexing logic suitable for
    sequences.
    """
    currentSortColumn = None
    def __init__(self, webTranslator, columns, defaultSortColumn,
                 defaultSortAscending):
        self.webTranslator = webTranslator
        self.columns = {}
        self.columnNames = []
        for col in columns:
            # see comment in TimestampAttributeColumn
            if isinstance(col, timestamp):
                col = TimestampAttributeColumn(col)
            else:
                col = IColumn(col)

            if defaultSortColumn is None:
                defaultSortColumn = col.sortAttribute()
            if (defaultSortColumn is not None
                and col.sortAttribute() is defaultSortColumn):
                self.currentSortColumn = col

            attributeID = unicode(col.attributeID, 'ascii')
            self.columns[attributeID] = col
            self.columnNames.append(attributeID)
        self.isAscending = defaultSortAscending
        if self.currentSortColumn is None:
            self._cannotDetermineSort(defaultSortColumn)


    def _cannotDetermineSort(self, defaultSortColumn):
        """
        This is an internal method designed to be overridden *only* by the classes
        in this module.

        It will be called if:

            * No explicit sort column was specified, and none of the columns
              specified is sortable, or,

            * An explicit sort column was specified, but it is not present in
              the list of columns specified.

        In other words, this method is called when there is no consistent sort
        that can be performed based on the caller's input to the constructor.
        In the default case of the inequality-based model, nothing can be done
        about this, and an exception is raised.  In the old index-based
        scrolltable, certain implicit sorts will appear to work, so those
        continue to work for those tables.  However, Users are advised to avoid
        the index-based scrolltable in general, and this subtly broken implicit
        behavior specifically.

        @param defaultSortColumn: something adaptable to L{IColumn}, or None,
        which subclasses may use to accept a sort that is not related to any
        extant columns in this table.

        @raise: L{Unsortable}, always.  Some subclasses can deal with this case
        better.
        """
        raise Unsortable('you must provide a sortable column')


    def resort(self, columnName):
        """
        Re-sort the table.

        @param columnName: the name of the column to sort by.  This is a string
        because it is passed from the browser.
        """
        csc = self.currentSortColumn
        newSortColumn = self.columns[columnName]
        if newSortColumn is None:
            raise Unsortable('column %r has no sort attribute' % (columnName,))
        if csc is newSortColumn:
            self.isAscending = not self.isAscending
        else:
            self.currentSortColumn = newSortColumn
            self.isAscending = True
        return self.isAscending
    expose(resort)


    def linkToItem(self, item):
        """
        Return a URL that the row for C{item} should link to, by asking the
        L{xmantissa.ixmantissa.IWebTranslator} in C{self.store}

        @return: C{unicode} URL
        """
        return unicode(self.webTranslator.toWebID(item), 'ascii')


    def itemFromLink(self, link):
        """
        Inverse of L{linkToItem}.

        @rtype: L{axiom.item.Item}
        """
        return self.webTranslator.fromWebID(link)



def _webTranslator(store, fallback):
    """
    Discover a web translator based on an Axiom store and a specified default.
    Prefer the specified default.

    This is an implementation detail of various initializers in this module
    which require an L{IWebTranslator} provider.  Some of those initializers
    did not previously require a webTranslator, so this function will issue a
    L{UserWarning} if no L{IWebTranslator} powerup exists for the given store
    and no fallback is provided.

    @param store: an L{axiom.store.Store}
    @param fallback: a provider of L{IWebTranslator}, or None

    @return: 'fallback', if it is provided, or the L{IWebTranslator} powerup on
    'store'.
    """
    if fallback is None:
        fallback = IWebTranslator(store, None)
        if fallback is None:
            warnings.warn(
                "No IWebTranslator plugin when creating Scrolltable - broken "
                "configuration, now deprecated!  Try passing webTranslator "
                "keyword argument.", category=DeprecationWarning,
                stacklevel=4)
    return fallback



class InequalityModel(_ScrollableBase):
    """
    This is a utility base class for things which want to communicate about
    large sets of Axiom items with a remote client.

    The first such implementation is L{ScrollingElement}, which communicates
    with a Nevow Athena JavaScript client.

    @ivar webTranslator: A L{IWebTranslator} provider for resolving and
    creating web links for items.

    @ivar columns: A mapping of attribute identifiers to L{IColumn}
    providers.

    @ivar columnNames: A list of attribute identifiers.

    @ivar isAscending: A boolean indicating the current order of the sort.

    @ivar currentSortColumn: An L{IColumn} representing the current
    sort key.
    """

    def __init__(self, store, itemType, baseConstraint, columns,
                 defaultSortColumn, defaultSortAscending,
                 webTranslator=None):
        """
        Create a new InequalityModel.

        @param store: the store to perform queries against when the client asks
        for data.
        @type store: L{axiom.store.Store}

        @param itemType: the type of item that will be returned by this
        L{InequalityModel}.
        @type itemType: a subclass of L{axiom.item.Item}.

        @param baseConstraint: an L{IQuery} provider that specifies the set of
        rows within this model.

        @param columns: a list of L{IColumn} providers listing the columns to
        be sent to the client.

        @param defaultSortColumn: an element of the C{columns} argument to sort
        by, by default.

        @param defaultSortAscending: is the sort ascending?  XXX: this is not
        implemented and may not even be a good idea to implement.

        @param webTranslator: an L{IWebTranslator} provider used to generate
        IDs for the client from the C{itemType}'s storeID.
        """
        super(InequalityModel, self).__init__(
            _webTranslator(store, webTranslator), columns,
             defaultSortColumn, defaultSortAscending)
        self.store = store
        self.itemType = itemType
        self.baseConstraint = baseConstraint


    def inequalityQuery(self, constraint, count, isAscending):
        """
        Perform a query to obtain some rows from the table represented
        by this model, at the behest of a networked client.

        @param constraint: an additional constraint to apply to the
        query.
        @type constraint: L{axiom.iaxiom.IComparison}.

        @param count: the maximum number of rows to return.
        @type count: C{int}

        @param isAscending: a boolean describing whether the query
        should be yielding ascending or descending results.
        @type isAscending: C{bool}

        @return: an query which will yield some results from this
        model.
        @rtype: L{axiom.iaxiom.IQuery}
        """
        if self.baseConstraint is not None:
            if constraint is not None:
                constraint = AND(self.baseConstraint, constraint)
            else:
                constraint = self.baseConstraint
        # build the sort
        currentSortAttribute = self.currentSortColumn.sortAttribute()
        if isAscending:
            sort = (currentSortAttribute.ascending,
                    self.itemType.storeID.ascending)
        else:
            sort = (currentSortAttribute.descending,
                    self.itemType.storeID.descending)
        return self.store.query(self.itemType, constraint, sort=sort,
                                limit=count).distinct()


    def rowsAfterValue(self, value, count):
        """
        Retrieve some rows at or after a given sort-column value.

        @param value: Starting value in the index for the current sort column
        at which to start returning results.  Rows with a column value for the
        current sort column which is greater than or equal to this value will
        be returned.

        @type value: Some type compatible with the current sort column, or
        None, to specify the beginning of the data.

        @param count: The maximum number of rows to return.
        @type count: C{int}

        @return: A list of row data, ordered by the current sort column,
        beginning at C{value} and containing at most C{count} elements.
        """
        if value is None:
            query = self.inequalityQuery(None, count, True)
        else:
            pyvalue = self._toComparableValue(value)
            currentSortAttribute = self.currentSortColumn.sortAttribute()
            query = self.inequalityQuery(currentSortAttribute >= pyvalue, count, True)
        return self.constructRows(query)
    expose(rowsAfterValue)


    def rowsAfterItem(self, item, count):
        """
        Retrieve some rows after a given item, not including the given item.

        @param item: then L{Item} to request something after.
        @type item: this L{InequalityModel}'s L{itemType} attribute.

        @param count: The maximum number of rows to return.
        @type count: L{int}

        @return: A list of row data, ordered by the current sort column,
        beginning immediately after C{item}.
        """
        currentSortAttribute = self.currentSortColumn.sortAttribute()
        value = currentSortAttribute.__get__(item, type(item))
        firstQuery = self.inequalityQuery(
            AND(currentSortAttribute == value,
                self.itemType.storeID > item.storeID),
            count, True)
        results = self.constructRows(firstQuery)
        count -= len(results)
        if count:
            secondQuery = self.inequalityQuery(
                currentSortAttribute > value,
                count, True)
            results.extend(self.constructRows(secondQuery))
        return results


    def rowsAfterRow(self, rowObject, count):
        """
        Wrapper around L{rowsAfterItem} which accepts the web ID for a item
        instead of the item itself.

        @param rowObject: a dictionary mapping strings to column values, sent
        from the client.  One of those column values must be C{__id__} to
        uniquely identify a row.

        @param count: an integer, the number of rows to return.
        """
        webID = rowObject['__id__']
        return self.rowsAfterItem(
            self.webTranslator.fromWebID(webID),
            count)
    expose(rowsAfterRow)


    def rowsBeforeRow(self, rowObject, count):
        """
        Wrapper around L{rowsBeforeItem} which accepts the web ID for a item
        instead of the item itself.

        @param rowObject: a dictionary mapping strings to column values, sent
        from the client.  One of those column values must be C{__id__} to
        uniquely identify a row.

        @param count: an integer, the number of rows to return.
        """
        webID = rowObject['__id__']
        return self.rowsBeforeItem(
            self.webTranslator.fromWebID(webID),
            count)
    expose(rowsBeforeRow)


    def _toComparableValue(self, value):
        """
        Trivial wrapper which takes into account the possibility that our sort
        column might not have defined the C{toComparableValue} method.

        This can probably serve as a good generic template for some
        infrastructure to deal with arbitrarily-potentially-missing methods
        from certain versions of interfaces, but we didn't take it any further
        than it needed to go for this system's fairly meagre requirements.
        *Please* feel free to refactor upwards as necessary.
        """
        if hasattr(self.currentSortColumn, 'toComparableValue'):
            return self.currentSortColumn.toComparableValue(value)
        # Retrieve the location of the class's definition so that we can alert
        # the user as to where they need to insert their implementation.
        classDef = self.currentSortColumn.__class__
        filename = inspect.getsourcefile(classDef)
        lineno = inspect.findsource(classDef)[1]
        warnings.warn_explicit(
            "IColumn implementor " + qual(self.currentSortColumn.__class__) + " "
            "does not implement method toComparableValue.  This is required since "
            "Mantissa 0.6.6.",
            DeprecationWarning, filename, lineno)
        return value


    def rowsBeforeValue(self, value, count):
        """
        Retrieve display data for rows with sort-column values less than the
        given value.

        @type value: Some type compatible with the current sort column.
        @param value: Starting value in the index for the current sort column
        at which to start returning results.  Rows with a column value for the
        current sort column which is less than this value will be returned.

        @type count: C{int}
        @param count: The number of rows to return.

        @return: A list of row data, ordered by the current sort column, ending
        at C{value} and containing at most C{count} elements.
        """
        if value is None:
            query = self.inequalityQuery(None, count, False)
        else:
            pyvalue = self._toComparableValue(value)
            currentSortAttribute = self.currentSortColumn.sortAttribute()
            query = self.inequalityQuery(
                currentSortAttribute < pyvalue, count, False)
        return self.constructRows(query)[::-1]
    expose(rowsBeforeValue)


    def rowsBeforeItem(self, item, count):
        """
        The inverse of rowsAfterItem.

        @param item: then L{Item} to request rows before.
        @type item: this L{InequalityModel}'s L{itemType} attribute.

        @param count: The maximum number of rows to return.
        @type count: L{int}

        @return: A list of row data, ordered by the current sort column,
        beginning immediately after C{item}.
        """
        currentSortAttribute = self.currentSortColumn.sortAttribute()
        value = currentSortAttribute.__get__(item, type(item))
        firstQuery = self.inequalityQuery(
            AND(currentSortAttribute == value,
                self.itemType.storeID < item.storeID),
            count, False)
        results = self.constructRows(firstQuery)
        count -= len(results)
        if count:
            secondQuery = self.inequalityQuery(currentSortAttribute < value,
                                               count, False)
            results.extend(self.constructRows(secondQuery))
        return results[::-1]



class IndexingModel(_ScrollableBase):
    """
    Mixin for "model" implementations of an in-browser scrollable list of
    elements.

    @ivar webTranslator: A L{IWebTranslator} provider for resolving and
    creating web links for items.

    @ivar columns: A mapping of attribute identifiers to L{IColumn}
    providers.

    @ivar columnNames: A list of attribute identifiers.

    @ivar isAscending: A boolean indicating the current order of the sort.

    @ivar currentSortColumn: An L{IColumn} representing the current
    sort key.
    """
    def requestRowRange(self, rangeBegin, rangeEnd):
        """
        Retrieve display data for the given range of rows.

        @type rangeBegin: C{int}
        @param rangeBegin: The index of the first row to retrieve.

        @type rangeEnd: C{int}
        @param rangeEnd: The index of the last row to retrieve.

        @return: A C{list} of C{dict}s giving row data.
        """
        return self.constructRows(self.performQuery(rangeBegin, rangeEnd))
    expose(requestRowRange)


    # The rest takes care of responding to requests from the client.
    def getTableMetadata(self):
        """
        Retrieve a description of the various properties of this scrolltable.

        @return: A sequence containing 5 elements.  They are, in order, a
        list of the names of the columns present, a mapping of column names
        to two-tuples of their type and a boolean indicating their
        sortability, the total number of rows in the scrolltable, the name
        of the default sort column, and a boolean indicating whether or not
        the current sort order is ascending.
        """
        coltypes = {}
        for (colname, column) in self.columns.iteritems():
            sortable = column.sortAttribute() is not None
            coltype = column.getType()
            if coltype is not None:
                coltype = unicode(coltype, 'ascii')
            coltypes[colname] = (coltype, sortable)

        if self.currentSortColumn:
            csc = unicode(self.currentSortColumn.sortAttribute().attrname, 'ascii')
        else:
            csc = None

        return [self.columnNames, coltypes, self.requestCurrentSize(),
                csc, self.isAscending]
    expose(getTableMetadata)


    def requestCurrentSize(self):
        return self.performCount()
    expose(requestCurrentSize)


    def performAction(self, name, rowID):
        method = getattr(self, 'action_' + name)
        item = self.itemFromLink(rowID)
        return method(item)
    expose(performAction)


    # Override these two in a subclass
    def performCount(self):
        """
        Override this in a subclass.

        @rtype: C{int}
        @return: The total number of elements in this scrollable.
        """
        raise NotImplementedError()


    def performQuery(self, rangeBegin, rangeEnd):
        """
        Override this in a subclass.

        @rtype: C{list}
        @return: Elements from C{rangeBegin} to C{rangeEnd} of the
        underlying data set, as ordered by the value of
        C{currentSortColumn} sort column in the order indicated by
        C{isAscending}.
        """
        raise NotImplementedError()

Scrollable = IndexingModel      # The old (deprecated) name of this class.

class ScrollableView(object):
    """
    Mixin for structuring model data in the way expected by
    Mantissa.ScrollTable.ScrollingWidget.

    Subclasses must also mix in L{_ScrollableBase} to provide required attributes
    and methods.
    """

    jsClass = u'Mantissa.ScrollTable.ScrollingWidget'
    fragmentName = 'scroller'

    def constructRows(self, items):
        """
        Build row objects that are serializable using Athena for sending to the
        client.

        @param items: an iterable of objects compatible with my columns'
        C{extractValue} methods.

        @return: a list of dictionaries, where each dictionary has a string key
        for each column name in my list of columns.
        """
        rows = []
        for item in items:
            row = dict((colname, col.extractValue(self, item))
                       for (colname, col) in self.columns.iteritems())
            link = self.linkToItem(item)
            if link is not None:
                row[u'__id__'] = link
            rows.append(row)

        return rows



class ItemQueryScrollingFragment(IndexingModel, ScrollableView, LiveElement):
    """
    An L{ItemQueryScrollingFragment} is an Athena L{LiveElement} that can
    display an Axiom query using an inefficient, but precise, method for
    counting rows and getting data at given offsets when requested.

    New code which wants to display a scrollable list of data should probably
    use L{ScrollingElement} instead.
    """
    def __init__(self, store, itemType, baseConstraint, columns,
                 defaultSortColumn=None, defaultSortAscending=True,
                 webTranslator=None,
                 *a, **kw):
        self.store = store
        self.itemType = itemType
        self.baseConstraint = baseConstraint
        IndexingModel.__init__(
            self,
            _webTranslator(store, webTranslator),
            columns,
            defaultSortColumn,
            defaultSortAscending)
        LiveElement.__init__(self, *a, **kw)


    def _cannotDetermineSort(self, defaultSortColumn):
        """
        In this old, index-based way of doing things, we can do even more implicit
        horrible stuff to determine the sort column, or even give up completely
        and accept an implicit sort.  NB: this is still terrible behavior, but
        lots of old code relied on it, and since this class is legacy anyway it
        won't be deprecated or removed.

        We can also accept a sort column in this table that is not actually
        displayed or sent to the client at all.
        """
        if defaultSortColumn is not None:
            self.currentSortColumn = IColumn(defaultSortColumn)


    def getInitialArguments(self):
        return [self.getTableMetadata()]


    def performCount(self):
        return self.store.query(self.itemType, self.baseConstraint).count()


    def performQuery(self, rangeBegin, rangeEnd):
        if self.isAscending:
            sort = self.currentSortColumn.sortAttribute().ascending
        else:
            sort = self.currentSortColumn.sortAttribute().descending
        return list(self.store.query(self.itemType,
                                     self.baseConstraint,
                                     offset=rangeBegin,
                                     limit=rangeEnd - rangeBegin,
                                     sort=sort))
ScrollingFragment = ItemQueryScrollingFragment



class SequenceScrollingFragment(IndexingModel, ScrollableView, LiveElement):
    """
    Scrolltable implementation backed by any Python L{axiom.item.Item}
    sequence.
    """
    def __init__(self, store, elements, columns,
                 defaultSortColumn=None,
                 defaultSortAscending=True,
                 webTranslator=None, *a, **kw):
        IndexingModel.__init__(
            self,
            _webTranslator(store, webTranslator),
            columns,
            defaultSortColumn,
            defaultSortAscending)

        LiveElement.__init__(self, *a, **kw)
        self.store = store
        self.elements = elements


    def _cannotDetermineSort(self, defaultSortColumn):
        """
        Since this model back-ends to a fixed-index sequence anyway, we won't be
        sorting by anything and the default sort column will always be None.
        """


    def getInitialArguments(self):
        return [self.getTableMetadata()]


    def performCount(self):
        return len(self.elements)


    def performQuery(self, rangeBegin, rangeEnd):
        step = 1
        if not self.isAscending:
            # The ranges are from the end, not the beginning.
            rangeBegin = max(0, len(self.elements) - rangeBegin - 1)

            # Python is so very very confusing:
            # s[1:0:-1] == []
            # s[1:None:-1] == [s[0]]
            # s[1:-1:-1] == some crazy thing you don't even want to know
            rangeEnd = max(-1, len(self.elements) - rangeEnd - 1)
            if rangeEnd == -1:
                rangeEnd = None
            step = -1
        result = self.elements[rangeBegin:rangeEnd:step]
        return result



class StoreIDSequenceScrollingFragment(SequenceScrollingFragment):
    """
    Scrolltable implementation like L{SequenceScrollingFragment} but which is
    backed by a sequence of Item storeID values rather than Items themselves.
    """
    def performQuery(self, rangeBegin, rangeEnd):
        return map(
            self.store.getItemByID,
            super(
                StoreIDSequenceScrollingFragment,
                self).performQuery(rangeBegin, rangeEnd))


class SearchResultScrollingFragment(SequenceScrollingFragment):
    """
    Scrolltable implementation like L{SequenceScrollingFragment} but which is
    backed by a sequence of L{_PyLuceneHitWrapper} instances.

    XXX _PyLuceneHitWrapper should probably implement IFulltextIndexable instead
    of a subtly different interface.
    """
    def performQuery(self, rangeBegin, rangeEnd):
        results = SequenceScrollingFragment.performQuery(
            self, rangeBegin, rangeEnd)
        return [
            self.store.getItemByID(int(hit.uniqueIdentifier))
            for hit
            in results]



class ScrollingElement(InequalityModel, ScrollableView, LiveElement):
    """
    Element for scrolling lists of items, which uses L{InequalityModel}.
    """
    jsClass = u'Mantissa.ScrollTable.ScrollTable'
    fragmentName = 'inequality-scroller'

    def __init__(self, store, itemType, baseConstraint, columns,
                 defaultSortColumn=None, defaultSortAscending=True,
                 webTranslator=None,
                 *a, **kw):
        InequalityModel.__init__(
            self, store, itemType, baseConstraint, columns,
            defaultSortColumn, defaultSortAscending, webTranslator)
        LiveElement.__init__(self, *a, **kw)


    def _getColumnList(self):
        """
        Get a list of serializable objects that describe the interesting
        columns on our item type.  Columns which report having no type will be
        treated as having the type I{text}.

        @rtype: C{list} of C{dict}
        """
        columnList = []
        for columnName in self.columnNames:
            column = self.columns[columnName]
            type = column.getType()
            if type is None:
                type = 'text'
            columnList.append(
                {u'name': columnName,
                 u'type': type.decode('ascii')})
        return columnList


    def getInitialArguments(self):
        """
        Return the constructor arguments required for the JavaScript client class,
        Mantissa.ScrollTable.ScrollTable.

        @return: a 3-tuple of::

          - The unicode attribute ID of my current sort column
          - A list of dictionaries with 'name' and 'type' keys which are
            strings describing the name and type of all the columns in this
            table.
          - A bool indicating whether the sort direction is initially
            ascending.
        """
        ic = IColumn(self.currentSortColumn)
        return [ic.attributeID.decode('ascii'),
                self._getColumnList(),
                self.isAscending]




"""
Integration tests having primarily to do with Mantissa's HTTP features.
"""

from zope.interface import Interface, Attribute, implements

from twisted.python.reflect import qual
from twisted.python.filepath import FilePath
from twisted.python.components import registerAdapter
from twisted.internet.defer import Deferred
from twisted.trial.unittest import TestCase
from twisted.web import http

from nevow.inevow import ICurrentSegments, IRemainingSegments
from nevow.context import RequestContext
from nevow.url import URL
from nevow.testutil import renderPage, FakeRequest
from nevow.page import Element, renderer
from nevow.loaders import stan
from nevow.tags import directive
from nevow.guard import SESSION_KEY, GuardSession
from nevow import athena

from axiom.store import Store
from axiom.attributes import text
from axiom.item import Item
from axiom.dependency import installOn
from axiom.userbase import LoginSystem
from axiom.plugins.mantissacmd import Mantissa

import xmantissa
from xmantissa.ixmantissa import (
    INavigableFragment, IOfferingTechnician, IPreferenceAggregator,
    IWebTranslator)
from xmantissa.port import SSLPort
from xmantissa.offering import Offering
from xmantissa.product import Product
from xmantissa.web import SiteConfiguration
from xmantissa.webapp import PrivateApplication
from xmantissa.sharing import getEveryoneRole
from xmantissa.websharing import addDefaultShareID
from xmantissa.signup import UserInfoSignup


def getResource(site, uri, headers={}, cookies={}):
    """
    Retrieve the resource at the given URI from C{site}.

    Return a L{Deferred} which is called back with the request after
    resource traversal and rendering has finished.

    @type site: L{NevowSite}
    @param site: The site object from which to retrieve the resource.

    @type uri: C{str}
    @param uri: The absolute path to the resource to retrieve, eg
        I{/private/12345}.

    @type headers: C{dict}
    @param headers: HTTP headers to include in the request.
    """
    headers = headers.copy()
    cookies = cookies.copy()

    url = URL.fromString(uri)
    path = '/' + url.path

    args = {}
    for (k, v) in url.queryList():
        args.setdefault(k, []).append(v)
    remainingSegments = tuple(url.pathList())
    request = FakeRequest(
        isSecure=url.scheme == 'https', uri=path, args=args, headers=headers,
        cookies=cookies, currentSegments=())
    requestContext = RequestContext(parent=site.context, tag=request)
    requestContext.remember((), ICurrentSegments)
    requestContext.remember(remainingSegments, IRemainingSegments)

    page = site.getPageContextForRequestContext(requestContext)
    page.addCallback(
        renderPage,
        topLevelContext=lambda tag: tag,
        reqFactory=lambda: request)
    page.addCallback(lambda ignored: request)
    return page


def getWithSession(site, redirectLimit, uri, headers={}, cookies={}):
    """
    Retrieve the resource at the given URI from C{site} while supplying a
    cookie identifying an existing session.

    @see L{getResource}
    """
    visited = []
    cookies = cookies.copy()

    result = Deferred()
    def rendered(request):
        if request.redirected_to is None:
            result.callback(request)
        else:
            visited.append(request.redirected_to)

            if visited.index(request.redirected_to) != len(visited) - 1:
                visited.append(request.redirected_to)
                result.errback(Exception("Redirect loop: %r" % (visited,)))
            elif len(visited) > redirectLimit:
                result.errback(Exception("Too many redirects: %r" % (visited,)))
            else:
                newHeaders = headers.copy()

                # Respect redirects
                location = URL.fromString(request.redirected_to)
                newHeaders['host'] = location.netloc

                # Respect cookies
                cookies.update(request.cookies)

                # str(URL) shouldn't really do what it does.
                page = getResource(
                    site, str(location), newHeaders, cookies)
                page.addCallbacks(rendered, result.errback)

    try:
        page = getResource(site, uri, headers, cookies)
    except:
        result.errback()
    else:
        page.addCallbacks(rendered, result.errback)
    return result



class IDummy(Interface):
    """
    An interface for which dummy items can be shared.
    """
    markup = Attribute(
        """
        The precise result to produce when rendering this object.
        """)



class DummyItem(Item):
    """
    An item which can be shared in order to test web sharing interactions.
    """
    implements(IDummy)

    markup = text(
        doc="""
        Some text to emit when rendering this item.
        """)



class DummyView(Element):
    """
    View for any L{IDummy}.
    """
    docFactory = stan(directive('display'))
    def __init__(self, dummy):
        Element.__init__(self)
        self.dummy = dummy

    @renderer
    def display(self, request, tag):
        return self.dummy.markup.encode('ascii')

registerAdapter(DummyView, IDummy, INavigableFragment)



class IntegrationTestsMixin:
    """
    L{TestCase} mixin defining setup and teardown such that requests can be
    made against a site strongly resembling an actual one.

    @type store: L{Store}
    @ivar store: The site store.

    @type web: L{WebSite}
    @ivar web: The site store's web site.

    @type login: L{LoginSystem}
    @ivar login: The site store's login system.

    @ivar site: A protocol factory created by the site store's L{WebSite}.
        This is probably a L{NevowSite}, but that should be an irrelevant
        detail.

    @type domain: C{unicode}
    @ivar domain: The canonical name of the website and the domain part used
        when creating users.
    """
    domain = u'example.com'

    def setUp(self):
        """
        Create a L{Store} with a L{WebSite} in it.  Get a protocol factory from
        the website and save it for tests to use.  Patch L{twisted.web.http}
        and L{nevow.guard} so that they don't create garbage timed calls that
        have to be cleaned up.
        """
        self.store = Store(filesdir=self.mktemp()) # See #2484
        Mantissa().installSite(self.store, self.domain, u'', False) # See #2483
        self.site = self.store.findUnique(SiteConfiguration)
        self.login = self.store.findUnique(LoginSystem)

        # Ports should be offering installation parameters.  This assumes a
        # TCPPort and an SSLPort are created by Mantissa.installSite. See
        # #538.  -exarkun
        self.store.query(
            SSLPort, SSLPort.factory == self.site).deleteFromStore()

        self.factory = self.site.getFactory()

        self.origFunctions = (http._logDateTimeStart,
                              GuardSession.checkExpired.im_func,
                              athena.ReliableMessageDelivery)
        http._logDateTimeStart = lambda: None
        GuardSession.checkExpired = lambda self: None
        athena.ReliableMessageDelivery = lambda *a, **kw: None


    def tearDown(self):
        """
        Restore the patched functions to their original state.
        """
        http._logDateTimeStart = self.origFunctions[0]
        GuardSession.checkExpired = self.origFunctions[1]
        athena.ReliableMessageDelivery = self.origFunctions[2]
        del self.origFunctions



class AnonymousWebSiteIntegrationTests(IntegrationTestsMixin, TestCase):
    """
    Integration (ie, not unit) tests for an anonymous user's interactions with
    a Mantissa L{WebSite}.
    """
    def _verifyResource(self, uri, verifyCallback):
        """
        Request the given URI and call the given callback with the resulting
        request.

        @type uri: C{str}

        @return: A L{Deferred} which will be called back with the result of
            C{verifyCallback} or which will errback if there is a problem
            requesting the resource or if the C{verifyCallback} raises an
            exception.
        """
        page = getResource(
            self.factory, uri, {'host': self.domain.encode('ascii')})
        page.addCallback(verifyCallback)
        return page


    def test_rootResource(self):
        """
        A sessionless, unauthenticated request for C{/} is responded to with a
        redirect to negotiate a session.
        """
        def rendered(request):
            redirectLocation = URL.fromString(request.redirected_to)
            key, path = redirectLocation.pathList()
            self.assertTrue(key.startswith(SESSION_KEY))
            self.assertEqual(path, '')
        return self._verifyResource('/', rendered)


    def test_mantissaStylesheet(self):
        """
        A sessionless, unauthenticated request for C{/Mantissa/mantissa.css} is
        responded to with the contents of the mantissa css file.
        """
        def rendered(request):
            staticPath = FilePath(xmantissa.__file__).sibling('static')
            self.assertEqual(
                request.accumulator,
                staticPath.child('mantissa.css').getContent())
        return self._verifyResource('/Mantissa/mantissa.css', rendered)


    def _fakeOfferings(self, store, offerings):
        """
        Override the adaption hook on the given L{Store} instance so that
        adapting it to L{IOfferingTechnician} returns an object which
        reports the given offerings as installed.
        """
        class FakeOfferingTechnician(object):
            def getInstalledOfferings(self):
                return offerings
        store.inMemoryPowerUp(FakeOfferingTechnician(), IOfferingTechnician)


    def test_offeringWithStaticContent(self):
        """
        L{StaticContent} has a L{File} child with the name of one of the
        offerings passed to its initializer which has a static content path.
        """
        # Make a fake offering to get its static content rendered.
        offeringName = u'name of the offering'
        offeringPath = FilePath(self.mktemp())
        offeringPath.makedirs()
        childName = 'content'
        childContent = 'the content'
        offeringPath.child(childName).setContent(childContent)

        self._fakeOfferings(self.store, {
                offeringName: Offering(
                    offeringName, None, None, None, None, None, None,
                    offeringPath, None)})

        def rendered(request):
            self.assertEqual(request.accumulator, childContent)
        return self._verifyResource('/static/%s/%s' % (
                offeringName.encode('ascii'), childName), rendered)


    def test_loginOverHTTP(self):
        """
        The login page is displayed over HTTP if there is no SSL server
        available.
        """
        page = getWithSession(
            self.factory, 2, '/login', {'host': self.domain.encode('ascii')})
        def rendered(request):
            # Sanity check.
            self.assertFalse(
                request.isSecure(), "Somehow managed to get HTTPS.")
            # This is a weak assertion.  There should be a better way to verify
            # that this is the LoginPage because that's what I care about - not
            # what's in the template. -exarkun
            self.assertIn("login-form", request.accumulator)
        page.addCallback(rendered)
        return page


    def test_loginOverHTTPS(self):
        """
        The login page is displayed over HTTPS even if it is initially
        requested over HTTP, if there is an SSL server available.
        """
        # Create the necessary SSL server
        SSLPort(store=self.store, factory=self.site, portNumber=443)

        page = getWithSession(
            self.factory, 3, '/login', {'host': self.domain.encode('ascii')})
        def rendered(request):
            self.assertTrue(request.isSecure(), "Page not served over HTTPS.")
            self.assertIn("login-form", request.accumulator)
        page.addCallback(rendered)
        return page


    def _createSignup(self, at):
        # Create a signup mechanism at the specified URL
        installOn(UserInfoSignup(store=self.store, prefixURL=at), self.store)

        # Ensure there will be a domain available for UserInfoSignup's view to
        # use.  This isn't really significant to the test, it's just necessary
        # with the implementation which exists as I am writing this
        # test. -exarkun
        self.login.addAccount(
            u'alice', u'example.com', u'password', internal=True)


    def test_userInfoSignupOverHTTP(self):
        """
        The L{UserInfoSignup} page is displayed over HTTP if there is no SSL
        server available.
        """
        self._createSignup(u'foobar-signup')

        # It is not necessarily the case that we want /signup to require a
        # session.  However, that is how it is currently implemented so I will
        # not change it now. -exarkun
        page = getWithSession(
            self.factory, 3, '/foobar-signup',
            {'host': self.domain.encode('ascii')})
        def rendered(request):
            # Sanity check.
            self.assertFalse(
                request.isSecure(), "Somehow managed to get HTTPS.")
            self.assertIn("Create Account", request.accumulator)
        page.addCallback(rendered)
        return page


    def test_userInfoSignupOverHTTPS(self):
        """
        The signup page is displayed over HTTPS even if it is initially
        requested over HTTP, if there is an SSL server available.
        """
        # Create the necessary SSL server
        SSLPort(store=self.store, factory=self.site, portNumber=443)

        self._createSignup(u'barbaz-signup')

        page = getWithSession(
            self.factory, 3, '/barbaz-signup',
            {'host': self.domain.encode('ascii')})
        def rendered(request):
            self.assertTrue(request.isSecure(), "Page not served over HTTPS.")
            self.assertIn("Create Account", request.accumulator)
        page.addCallback(rendered)
        return page


    def test_userSharedResource(self):
        """
        An item shared by a user to everybody can be accessed by an
        unauthenticated user.
        """
        # Make a user to own the shared item.
        username = u'alice'
        aliceAccount = self.login.addAccount(
            username, self.domain, u'password', internal=True)
        aliceStore = aliceAccount.avatars.open()

        # Make an item to share.
        sharedContent = u'content owned by alice and shared to everyone'
        shareID = getEveryoneRole(aliceStore).shareItem(
            DummyItem(store=aliceStore, markup=sharedContent)).shareID

        # Get it.
        page = getWithSession(
            self.factory, 2, '/users/%s/%s' % (
                username.encode('ascii'), shareID.encode('ascii')),
            {'host': self.domain.encode('ascii')})
        def rendered(request):
            self.assertIn(sharedContent.encode('ascii'), request.accumulator)
        page.addCallback(rendered)
        return page


    def test_defaultUserSharedResource(self):
        """
        The resource for the default share can be accessed by an
        unauthenticated user.
        """
        # Make a user to own the shared item.
        username = u'alice'
        aliceAccount = self.login.addAccount(
            username, self.domain, u'password', internal=True)
        aliceStore = aliceAccount.avatars.open()

        # Make an item to share.
        sharedContent = u'content owned by alice and shared to everyone'
        shareID = getEveryoneRole(aliceStore).shareItem(
            DummyItem(store=aliceStore, markup=sharedContent)).shareID

        # Make it the default share.
        addDefaultShareID(aliceStore, shareID, 0)

        # Get it.
        page = getWithSession(
            self.factory, 3, '/users/%s' % (username.encode('ascii'),),
            {'host': self.domain.encode('ascii')})
        def rendered(request):
            self.assertIn(sharedContent.encode('ascii'), request.accumulator)
        page.addCallback(rendered)
        return page



class AuthenticatedWebSiteIntegrationTests(IntegrationTestsMixin, TestCase):
    """
    Integration (ie, not unit) tests for an authenticated user's interactions
    with a Mantissa L{WebSite}.

    @type username: C{unicode}
    @ivar username: The localpart used when creating users.

    @type cookies: C{dict}
    @ivar cookies: The cookies to use in order to use the authenticated session
        created in L{setUp}.
    """
    username = u'alice'

    def setUp(self):
        """
        Create an account and log in using it.
        """
        IntegrationTestsMixin.setUp(self)

        # Make an account to be already logged in.
        self.userAccount = self.login.addAccount(
            self.username, self.domain, u'password', internal=True)
        self.userStore = self.userAccount.avatars.open()

        # Make a product that includes PrivateApplication.  This is probably
        # the minimum requirement for web access.
        web = Product(store=self.store,
                      types=[qual(PrivateApplication)])
        # Give it to Alice.
        web.installProductOn(self.userStore)

        # Log in to the web as Alice.
        login = getWithSession(
            self.factory, 3, '/__login__?username=%s@%s&password=%s' % (
                self.username.encode('ascii'), self.domain.encode('ascii'),
                'password'),
            {'host': self.domain.encode('ascii')})
        def loggedIn(request):
            self.cookies = request.cookies
        login.addCallback(loggedIn)
        return login


    def test_authenticatedResetPasswordRedirectsToSettings(self):
        """
        When a user is already logged in, navigating to C{/resetPassword}
        redirects them to their own settings page.
        """
        prefPage = IPreferenceAggregator(self.userStore)
        urlPath = IWebTranslator(self.userStore).linkTo(prefPage.storeID)

        # Get the password reset resource.
        page = getResource(
            self.factory, '/resetPassword',
            headers={'host': self.domain.encode('ascii')},
            cookies=self.cookies)

        def rendered(request):
            # Make sure it's a redirect to the settings page.
            self.assertEquals(
                'http://' + self.domain.encode('ascii') + urlPath,
                request.redirected_to)
        page.addCallback(rendered)
        return page


    def test_userSharedResource(self):
        """
        An item shared by a user to everybody can be accessed by that user.
        """
        # Make an item and share it.
        sharedContent = u'content owned by alice and shared to everyone'
        shareID = getEveryoneRole(self.userStore).shareItem(
            DummyItem(store=self.userStore, markup=sharedContent)).shareID

        page = getResource(
            self.factory, '/users/%s/%s' % (
                self.username.encode('ascii'), shareID.encode('ascii')),
            {'host': self.domain.encode('ascii')},
            self.cookies)

        def rendered(request):
            self.assertIn(sharedContent, request.accumulator)
        page.addCallback(rendered)
        return page


    def test_defaultUserSharedResource(self):
        """
        The resource for the default share can be accessed by an authenticated
        user.
        """
        # Make an item and share it.
        sharedContent = u'content owned by alice and shared to everyone'
        shareID = getEveryoneRole(self.userStore).shareItem(
            DummyItem(store=self.userStore, markup=sharedContent)).shareID

        # Make it the default.
        addDefaultShareID(self.userStore, shareID, 0)

        # Get it.
        page = getWithSession(
            self.factory, 1, '/users/%s' % (self.username.encode('ascii'),),
            {'host': self.domain.encode('ascii')},
            self.cookies)

        def rendered(request):
            self.assertIn(sharedContent, request.accumulator)
        page.addCallback(rendered)
        return page



class UserSubdomainWebSiteIntegrationTests(IntegrationTestsMixin, TestCase):
    """
    @type share: L{Share}
    @ivar share: The share for the shared item.

    @type sharedContent: C{unicode}
    @ivar sharedContent: The text which will appear in the view for C{share}.

    @type username: C{unicode}
    @ivar username: The localpart of the user which shared C{share}.

    @type virtualHost: C{unicode}
    @ivar virtualHost: The full domain name of a user-specific subdomain for
        the user which shared C{share}.
    """
    def setUp(self):
        """
        Create a user account and share an item from it to everyone.
        """
        IntegrationTestsMixin.setUp(self)

        # Make an account to go with this virtual host.
        self.username = u'alice'
        self.userAccount = self.login.addAccount(
            self.username, self.domain, u'password', internal=True)
        self.userStore = self.userAccount.avatars.open()
        self.virtualHost = u'.'.join((self.username, self.domain))

        # Share something that we'll try to load.
        self.sharedContent = u'content owned by alice and shared to everyone'
        self.share = getEveryoneRole(self.userStore).shareItem(
            DummyItem(store=self.userStore, markup=self.sharedContent))


    def test_anonymousUserVirtualHost(self):
        """
        A request by an anonymous user for I{/shareid} on a subdomain of the
        website's is responded to with the page for indicated item shared by
        the user to whom that subdomain corresponds.
        """
        page = getWithSession(
            self.factory, 2, '/' + self.share.shareID.encode('ascii'),
            {'host': self.virtualHost.encode('ascii')})
        def rendered(request):
            self.assertIn(
                self.sharedContent.encode('ascii'), request.accumulator)
        page.addCallback(rendered)
        return page


    def test_authenticatedUserVirtualHost(self):
        """
        A request by an authenticated user for I{/shareid} on a subdomain of
        the website's is responded to in the same way as the same request made
        by an anonymous user.
        """
        # Make an account as which to authenticate.
        username = u'bob'
        bobAccount = self.login.addAccount(
            username, self.domain, u'password', internal=True)
        bobStore = bobAccount.avatars.open()

        # Make a product that includes PrivateApplication.  This supposes that
        # viewing user-subdomain virtual hosting is the responsibility of
        # PrivateApplication.
        web = Product(store=self.store,
                      types=[qual(PrivateApplication)])
        # Give it to Bob.
        web.installProductOn(bobStore)

        # Log in through the web as Bob.
        cookies = {}
        login = getWithSession(
            self.factory, 3, '/__login__?username=%s@%s&password=%s' % (
                username.encode('ascii'), self.domain.encode('ascii'), 'password'),
            {'host': self.domain.encode('ascii')})
        def loggedIn(request):
            # Get the share page from the virtual host as the authenticated
            # user.
            return getResource(
                self.factory, '/' + self.share.shareID.encode('ascii'),
                headers={'host': self.virtualHost.encode('ascii')},
                cookies=request.cookies)
        login.addCallback(loggedIn)

        def rendered(request):
            # Make sure we're really authenticated.
            self.assertIn(username.encode('ascii'), request.accumulator)
            # Make sure the shared thing is there.
            self.assertIn(
                self.sharedContent.encode('ascii'), request.accumulator)
        login.addCallback(rendered)
        return login


    def test_authenticatedUserVirtualHostDefaultShare(self):
        """
        A request by an authenticated user for I{/} on a subdomain of of the
        website's is responded to in the same way as the same request made by
        an anonymous user.
        """
        # Make an account as which to authenticate.
        username = u'bob'
        bobAccount = self.login.addAccount(
            username, self.domain, u'password', internal=True)
        bobStore = bobAccount.avatars.open()

        # Make a product that includes PrivateApplication.  This supposes that
        # viewing user-subdomain virtual hosting is the responsibility of
        # PrivateApplication.
        web = Product(store=self.store,
                      types=[qual(PrivateApplication)])
        # Give it to Bob.
        web.installProductOn(bobStore)

        # Make something the default share.
        addDefaultShareID(self.userStore, self.share.shareID, 0)

        # Log in through the web as Bob.
        cookies = {}
        login = getWithSession(
            self.factory, 3, '/__login__?username=%s@%s&password=%s' % (
                username.encode('ascii'), self.domain.encode('ascii'), 'password'),
            {'host': self.domain.encode('ascii')})
        def loggedIn(request):
            # Get the share page as the authenticated user.
            return getResource(
                self.factory, '/',
                headers={'host': self.virtualHost.encode('ascii')},
                cookies=request.cookies)
        login.addCallback(loggedIn)

        def rendered(request):
            # Make sure we're really authenticated.
            self.assertIn(username.encode('ascii'), request.accumulator)
            # Make sure the shared thing is there.
            self.assertIn(
                self.sharedContent.encode('ascii'), request.accumulator)
        login.addCallback(rendered)
        return login




from xmantissa import offering, stats
from xmantissa.webadmin import (TracebackViewer, LocalUserBrowser,
                                DeveloperApplication,
                                PortConfiguration)
from xmantissa.signup import SignupConfiguration

adminOffering = offering.Offering(
    name = u'mantissa',
    description = u'Powerups for administrative control of a Mantissa server.',
    siteRequirements = [],
    appPowerups = [stats.StatsService],
    installablePowerups = [("Signup Configuration", "Allows configuration of signup mechanisms", SignupConfiguration),
                           ("Traceback Viewer", "Allows viewing unhandled exceptions which occur on the server", TracebackViewer),
                           ("Port Configuration", "Allows manipulation of network service configuration.", PortConfiguration),
                           ("Local User Browser", "A page listing all users existing in this site's store.", LocalUserBrowser),
                           ("Admin REPL", "An interactive python prompt.", DeveloperApplication),
                           ("Offering Configuration", "Allows installation of Offerings on this site", offering.OfferingConfiguration),
                           ("Stats Observation", "Allows remote observation via AMP of gathered performance-related stats",
                            stats.RemoteStatsCollectorFactory),
                           ],
    loginInterfaces=(),
    themes = ())



# -*- test-case-name: xmantissa.test.test_website -*-

"""
This module defines the basic engine for web sites and applications using
Mantissa.  It defines the basic in-database web server, and an authentication
binding using nevow.guard.

To interact with the code defined here, create a web site using the
command-line 'axiomatic' program using the 'web' subcommand.
"""

import warnings

from zope.interface import implements

try:
    from cssutils import CSSParser
    CSSParser
except ImportError:
    CSSParser = None



from epsilon.structlike import record

from nevow.inevow import IRequest, IResource
from nevow.rend import Page, Fragment
from nevow import inevow
from nevow.static import File
from nevow.url import URL
from nevow import url
from nevow import athena

from axiom.iaxiom import IPowerupIndirector
from axiom import upgrade
from axiom.item import Item, _PowerupConnector, declareLegacyItem
from axiom.attributes import AND, integer, text, reference, bytes, boolean
from axiom.userbase import LoginSystem, getAccountNames
from axiom.dependency import installOn, uninstallFrom, installedOn

from xmantissa.ixmantissa import (
    ISiteRootPlugin, ISessionlessSiteRootPlugin,
    ISiteURLGenerator)
from xmantissa.port import TCPPort, SSLPort
from xmantissa.web import SiteConfiguration
from xmantissa.cachejs import theHashModuleProvider
from xmantissa._webutil import SiteRootMixin


class MantissaLivePage(athena.LivePage):
    """
    An L{athena.LivePage} which supports the global JavaScript modules
    collection that Mantissa provides as a root resource.

    All L{athena.LivePage} usages within and derived from Mantissa should
    subclass this.

    @ivar webSite: a L{WebSite} instance which provides site configuration
        information for generating links.

    @ivar hashCache: a cache which maps JS module names to
        L{xmantissa.cachejs.CachedJSModule} objects.
    @type hashCache: L{xmantissa.cachejs.HashedJSModuleProvider}

    @type _moduleRoot: L{URL}
    @ivar _moduleRoot: The base location for script tags which load Athena
        modules required by this page and widgets on this page.  This is set
        based on the I{Host} header in the request, so it is C{None} until
        the instance is actually rendered.
    """

    hashCache = theHashModuleProvider

    _moduleRoot = None

    def __init__(self, webSite, *a, **k):
        """
        Create a L{MantissaLivePage}.

        @param webSite: a L{WebSite} with a usable secure port implementation.
        """
        self.webSite = webSite
        athena.LivePage.__init__(self, transportRoot=url.root.child('live'),
                                 *a, **k)
        self._jsDepsMemo = self.hashCache.depsMemo


    def beforeRender(self, ctx):
        """
        Before rendering, retrieve the hostname from the request being
        responded to and generate an URL which will serve as the root for
        all JavaScript modules to be loaded.
        """
        request = IRequest(ctx)
        root = self.webSite.rootURL(request)
        self._moduleRoot = root.child('__jsmodule__')


    def getJSModuleURL(self, moduleName):
        """
        Retrieve an L{URL} object which references the given module name.

        This makes a 'best effort' guess as to an fully qualified HTTPS URL
        based on the hostname provided during rendering and the configuration
        of the site.  This is to avoid unnecessary duplicate retrieval of the
        same scripts from two different URLs by the browser.

        If such configuration does not exist, however, it will simply return an
        absolute path URL with no hostname or port.

        @raise NotImplementedError: if rendering has not begun yet and
        therefore beforeRender has not provided us with a usable hostname.
        """
        if self._moduleRoot is None:
            raise NotImplementedError(
                "JS module URLs cannot be requested before rendering.")
        moduleHash = self.hashCache.getModule(moduleName).hashValue
        return self._moduleRoot.child(moduleHash).child(moduleName)



JUST_SLASH = ('',)

class PrefixURLMixin(object):
    """
    Mixin for use by I[Sessionless]SiteRootPlugin implementors; provides a
    resourceFactory method which looks for an C{prefixURL} string on self,
    and calls and returns self.createResource().

    C{prefixURL} is a '/'-separated unicode string; it must be set before
    calling installOn.  To respond to the url C{http://example.com/foo/bar},
    use the prefixURL attribute u'foo/bar'.

    @ivar sessioned: Boolean indicating whether this object should
    powerup for L{ISiteRootPlugin}.  Note: this is only tested when
    L{installOn} is called.  If you change it later, it will have no
    impact.

    @ivar sessionless: Boolean indicating whether this object should
    powerup for ISessionlessSiteRootPlugin.  This is tested at the
    same time as L{sessioned}.
    """

    sessioned = False
    sessionless = False

    def __str__(self):
        return '/%s => item(%s)' % (self.prefixURL, self.__class__.__name__)


    def createResourceWith(self, webViewer):
        """
        Create and return an IResource.  This will only be invoked if
        the request matches the prefixURL specified on this object.
        May also return None to indicate that this object does not
        actually want to handle this request.

        Note that this will only be invoked for L{ISiteRootPlugin} powerups;
        L{ISessionlessSiteRootPlugin} powerups will only have C{createResource}
        invoked.
        """
        raise NotImplementedError(
            "PrefixURLMixin.createResourceWith(webViewer) should be "
            "implemented by subclasses (%r didn't)" % (
                self.__class__.__name__,))


    # ISiteRootPlugin
    def produceResource(self, request, segments, webViewer):
        """
        Return a C{(resource, subsegments)} tuple or None, depending on whether
        I wish to return an L{IResource} provider for the given set of segments
        or not.
        """
        def thunk():
            cr = getattr(self, 'createResource', None)
            if cr is not None:
                return cr()
            else:
                return self.createResourceWith(webViewer)
        return self._produceIt(segments, thunk)


    # ISessionlessSiteRootPlugin
    def sessionlessProduceResource(self, request, segments):
        """
        Return a C{(resource, subsegments)} tuple or None, depending on whether
        I wish to return an L{IResource} provider for the given set of segments
        or not.
        """
        return self._produceIt(segments, self.createResource)


    def _produceIt(self, segments, thunk):
        """
        Underlying implmeentation of L{PrefixURLMixin.produceResource} and
        L{PrefixURLMixin.sessionlessProduceResource}.

        @param segments: the URL segments to dispatch.

        @param thunk: a 0-argument callable which returns an L{IResource}
        provider, or None.

        @return: a 2-tuple of C{(resource, remainingSegments)}, or L{None}.
        """
        if not self.prefixURL:
            needle = ()
        else:
            needle = tuple(self.prefixURL.split('/'))
        S = len(needle)
        if segments[:S] == needle:
            if segments == JUST_SLASH:
                # I *HATE* THE WEB
                subsegments = segments
            else:
                subsegments = segments[S:]
            res = thunk()
            # Even though the URL matched up, sometimes we might still
            # decide to not handle this request (eg, some prerequisite
            # for our function is not met by the store).  Allow None
            # to be returned by createResource to indicate this case.
            if res is not None:
                return res, subsegments


    def __getPowerupInterfaces__(self, powerups):
        """
        Install me on something (probably a Store) that will be queried for
        ISiteRootPlugin providers.
        """

        #First, all the other powerups
        for x in powerups:
            yield x

        # Only 256 segments are allowed in URL paths.  We want to make sure
        # that static powerups always lose priority ordering to dynamic
        # powerups, since dynamic powerups will have information
        pURL = self.prefixURL
        priority = (pURL.count('/') - 256)
        if pURL == '':
            # Did I mention I hate the web?  Plugins at / are special in 2
            # ways.  Their segment length is kinda-sorta like 0 most of the
            # time, except when it isn't.  We subtract from the priority here
            # to make sure that [''] is lower-priority than ['foo'] even though
            # they are technically the same number of segments; the reason for
            # this is that / is special in that it pretends to be the parent of
            # everything and will score a hit for *any* URL in the hierarchy.
            # Above, we special-case JUST_SLASH to make sure that the other
            # half of this special-casing holds true.
            priority -= 1

        if not self.sessioned and not self.sessionless:
            warnings.warn(
                "Set either sessioned or sessionless on %r!  Falling back to "
                "deprecated providedBy() behavior" % (self.__class__.__name__,),
                DeprecationWarning,
                stacklevel=2)
            for iface in ISessionlessSiteRootPlugin, ISiteRootPlugin:
                if iface.providedBy(self):
                    yield (iface, priority)
        else:
            if self.sessioned:
                yield (ISiteRootPlugin, priority)
            if self.sessionless:
                yield (ISessionlessSiteRootPlugin, priority)


class StaticSite(PrefixURLMixin, Item):
    implements(ISessionlessSiteRootPlugin,     # implements both so that it
               ISiteRootPlugin)                # works in both super and sub
                                               # stores.
    typeName = 'static_web_site'
    schemaVersion = 2

    prefixURL = text()
    staticContentPath = text()

    sessioned = boolean(default=False)
    sessionless = boolean(default=True)

    def __str__(self):
        return '/%s => file(%s)' % (self.prefixURL, self.staticContentPath)

    def createResource(self):
        return File(self.staticContentPath)

    def installSite(self):
        """
        Not using the dependency system for this class because it's only
        installed via the command line, and multiple instances can be
        installed.
        """
        for iface, priority in self.__getPowerupInterfaces__([]):
            self.store.powerUp(self, iface, priority)

def upgradeStaticSite1To2(oldSite):
    newSite = oldSite.upgradeVersion(
        'static_web_site', 1, 2,
        staticContentPath=oldSite.staticContentPath,
        prefixURL=oldSite.prefixURL,
        sessionless=True)
    for pc in newSite.store.query(_PowerupConnector,
                                  AND(_PowerupConnector.powerup == newSite,
                                      _PowerupConnector.interface == u'xmantissa.ixmantissa.ISiteRootPlugin')):
        pc.item.powerDown(newSite, ISiteRootPlugin)
    return newSite
upgrade.registerUpgrader(upgradeStaticSite1To2, 'static_web_site', 1, 2)


class StaticRedirect(Item, PrefixURLMixin):
    implements(inevow.IResource,
               ISessionlessSiteRootPlugin,
               ISiteRootPlugin)

    schemaVersion = 2
    typeName = 'web_static_redirect'

    targetURL = text(allowNone=False)

    prefixURL = text(allowNone=False)

    sessioned = boolean(default=True)
    sessionless = boolean(default=True)

    def __str__(self):
        return '/%s => url(%s)' % (self.prefixURL, self.targetURL)

    def locateChild(self, ctx, segments):
        return self, ()

    def renderHTTP(self, ctx):
        return URL.fromContext(ctx).click(self.targetURL)

    def createResource(self):
        return self

def upgradeStaticRedirect1To2(oldRedirect):
    newRedirect = oldRedirect.upgradeVersion(
        'web_static_redirect', 1, 2,
        targetURL=oldRedirect.targetURL,
        prefixURL=oldRedirect.prefixURL)
    if newRedirect.prefixURL == u'':
        newRedirect.sessionless = False
        for pc in newRedirect.store.query(_PowerupConnector,
                                          AND(_PowerupConnector.powerup == newRedirect,
                                              _PowerupConnector.interface == u'xmantissa.ixmantissa.ISessionlessSiteRootPlugin')):
            pc.item.powerDown(newRedirect, ISessionlessSiteRootPlugin)
    return newRedirect
upgrade.registerUpgrader(upgradeStaticRedirect1To2, 'web_static_redirect', 1, 2)

class AxiomPage(Page):
    def renderHTTP(self, ctx):
        return self.store.transact(Page.renderHTTP, self, ctx)



class AxiomFragment(Fragment):
    def rend(self, ctx, data):
        return self.store.transact(Fragment.rend, self, ctx, data)



class StylesheetFactory(record('installedOfferingNames rootURL')):
    """
    Factory which creates resources for stylesheets which will rewrite URLs in
    them to be rooted at a particular location.

    @ivar installedOfferingNames: A C{list} of C{unicode} giving the names of
        the offerings which are installed and have a static content path.
        These are the offerings for which C{StaticContent} will find children,
        so these are the only offerings URLs pointed at which should be
        rewritten.

    @ivar rootURL: A one-argument callable which takes a request and returns an
        L{URL} which is to be used as the root of all URLs served by resources
        this factory creates.
    """
    def makeStylesheetResource(self, path, registry):
        """
        Return a resource for the css at the given path with its urls rewritten
        based on self.rootURL.
        """
        return StylesheetRewritingResourceWrapper(
            File(path), self.installedOfferingNames, self.rootURL)



class StylesheetRewritingResourceWrapper(
    record('resource installedOfferingNames rootURL')):
    """
    Resource which renders another resource using a request which rewrites CSS
    URLs.

    @ivar resource: Another L{IResource} which will be used to generate the
        response.

    @ivar installedOfferingNames: See L{StylesheetFactory.installedOfferingNames}

    @ivar rootURL: See L{StylesheetFactory.rootURL}
    """
    implements(IResource)

    def renderHTTP(self, context):
        """
        Render C{self.resource} through a L{StylesheetRewritingRequestWrapper}.
        """
        request = IRequest(context)
        request = StylesheetRewritingRequestWrapper(
            request, self.installedOfferingNames, self.rootURL)
        context.remember(request, IRequest)
        return self.resource.renderHTTP(context)



class StylesheetRewritingRequestWrapper(object):
    """
    Request which intercepts the response body, parses it as CSS, rewrites its
    URLs, and sends the serialized result.

    @ivar request: Another L{IRequest} object, methods of which will be used to
        implement this request.

    @ivar _buffer: A list of C{str} which have been passed to the write method.

    @ivar installedOfferingNames: See L{StylesheetFactory.installedOfferingNames}

    @ivar rootURL: See L{StylesheetFactory.rootURL}.
    """
    def __init__(self, request, installedOfferingNames, rootURL):
        self.request = request
        self._buffer = []
        self.installedOfferingNames = installedOfferingNames
        self.rootURL = rootURL


    def __getattr__(self, name):
        """
        Pass attribute lookups on to the wrapped request object.
        """
        return getattr(self.request, name)


    def write(self, bytes):
        """
        Buffer the given bytes for later processing.
        """
        self._buffer.append(bytes)


    def _replace(self, url):
        """
        Change URLs with absolute paths so they are rooted at the correct
        location.
        """
        segments = url.split('/')
        if segments[0] == '':
            root = self.rootURL(self.request)
            if segments[1] == 'Mantissa':
                root = root.child('static').child('mantissa-base')
                segments = segments[2:]
            elif segments[1] in self.installedOfferingNames:
                root = root.child('static').child(segments[1])
                segments = segments[2:]
            for seg in segments:
                root = root.child(seg)
            return str(root)
        return url


    def finish(self):
        """
        Parse the buffered response body, rewrite its URLs, write the result to
        the wrapped request, and finish the wrapped request.
        """
        stylesheet = ''.join(self._buffer)
        parser = CSSParser()
        css = parser.parseString(stylesheet)
        css.replaceUrls(self._replace)
        self.request.write(css.cssText)
        return self.request.finish()



class WebSite(Item, SiteRootMixin):
    """
    An IResource avatar which supports L{ISiteRootPlugin}s on user stores
    and a limited number of other statically defined children.
    """
    powerupInterfaces = (IResource,)
    implements(*powerupInterfaces + (IPowerupIndirector,))

    typeName = 'mantissa_web_powerup'
    schemaVersion = 6

    hitCount = integer(default=0)

    def cleartextRoot(self, hostname=None):
        """
        Return a string representing the HTTP URL which is at the root of this
        site.

        @param hostname: An optional unicode string which, if specified, will
        be used as the hostname in the resulting URL, regardless of the
        C{hostname} attribute of this item.
        """
        warnings.warn(
            "Use ISiteURLGenerator.rootURL instead of WebSite.cleartextRoot.",
            category=DeprecationWarning,
            stacklevel=2)
        if self.store.parent is not None:
            generator = ISiteURLGenerator(self.store.parent)
        else:
            generator = ISiteURLGenerator(self.store)
        return generator.cleartextRoot(hostname)


    def encryptedRoot(self, hostname=None):
        """
        Return a string representing the HTTPS URL which is at the root of this
        site.

        @param hostname: An optional unicode string which, if specified, will
        be used as the hostname in the resulting URL, regardless of the
        C{hostname} attribute of this item.
        """
        warnings.warn(
            "Use ISiteURLGenerator.rootURL instead of WebSite.encryptedRoot.",
            category=DeprecationWarning,
            stacklevel=2)
        if self.store.parent is not None:
            generator = ISiteURLGenerator(self.store.parent)
        else:
            generator = ISiteURLGenerator(self.store)
        return generator.encryptedRoot(hostname)


    def maybeEncryptedRoot(self, hostname=None):
        """
        Returning a string representing the HTTPS URL which is at the root of
        this site, falling back to HTTP if HTTPS service is not available.

        @param hostname: An optional unicode string which, if specified, will
        be used as the hostname in the resulting URL, regardless of the
        C{hostname} attribute of this item.
        """
        warnings.warn(
            "Use ISiteURLGenerator.rootURL instead of "
            "WebSite.maybeEncryptedRoot",
            category=DeprecationWarning,
            stacklevel=2)
        if self.store.parent is not None:
            generator = ISiteURLGenerator(self.store.parent)
        else:
            generator = ISiteURLGenerator(self.store)
        root = generator.encryptedRoot(hostname)
        if root is None:
            root = generator.cleartextRoot(hostname)
        return root


    def rootURL(self, request):
        """
        Simple utility function to provide a root URL for this website which is
        appropriate to use in links generated in response to the given request.

        @type request: L{twisted.web.http.Request}
        @param request: The request which is being responded to.

        @rtype: L{URL}
        @return: The location at which the root of the resource hierarchy for
            this website is available.
        """
        warnings.warn(
            "Use ISiteURLGenerator.rootURL instead of WebSite.rootURL.",
            category=DeprecationWarning,
            stacklevel=2)
        if self.store.parent is not None:
            generator = ISiteURLGenerator(self.store.parent)
        else:
            generator = ISiteURLGenerator(self.store)
        return generator.rootURL(request)


    def rootChild_resetPassword(self, req, webViewer):
        """
        Redirect authenticated users to their settings page (hopefully they
        have one) when they try to reset their password.

        This is the wrong way for this functionality to be implemented.  See
        #2524.
        """
        from xmantissa.ixmantissa import IWebTranslator, IPreferenceAggregator
        return URL.fromString(
            IWebTranslator(self.store).linkTo(
                IPreferenceAggregator(self.store).storeID))


    def setServiceParent(self, parent):
        """
        Compatibility hack necessary to prevent the Axiom service startup
        mechanism from barfing.  Even though this Item is no longer an IService
        powerup, it will still be found as one one more time and this method
        will be called on it.
        """


    def getFactory(self):
        """
        @see L{setServiceParent}.
        """
        return self.store.findUnique(SiteConfiguration).getFactory()


    def _getUsername(self):
        """
        Return a username, suitable for creating a L{VirtualHostWrapper} with.
        """
        return u'@'.join(getAccountNames(self.store).next())




class APIKey(Item):
    """
    Persistent record of a key used for accessing an external API.

    @cvar URCHIN: Constant name for the "Google Analytics" API
    (http://code.google.com/apis/maps/)
    @type URCHIN: C{unicode}
    """
    URCHIN = u'Google Analytics'

    apiName = text(
        doc="""
        L{APIKey} constant naming the API this key is for.
        """, allowNone=False)


    apiKey = text(
        doc="""
        The key.
        """, allowNone=False)


    def getKeyForAPI(cls, siteStore, apiName):
        """
        Get the API key for the named API, if one exists.

        @param siteStore: The site store.
        @type siteStore: L{axiom.store.Store}

        @param apiName: The name of the API.
        @type apiName: C{unicode} (L{APIKey} constant)

        @rtype: L{APIKey} or C{NoneType}
        """
        return siteStore.findUnique(
            cls, cls.apiName == apiName, default=None)
    getKeyForAPI = classmethod(getKeyForAPI)


    def setKeyForAPI(cls, siteStore, apiName, apiKey):
        """
        Set the API key for the named API, overwriting any existing key.

        @param siteStore: The site store to install the key in.
        @type siteStore: L{axiom.store.Store}

        @param apiName: The name of the API.
        @type apiName: C{unicode} (L{APIKey} constant)

        @param apiKey: The key for accessing the API.
        @type apiKey: C{unicode}

        @rtype: L{APIKey}
        """
        existingKey = cls.getKeyForAPI(siteStore, apiName)
        if existingKey is None:
            return cls(store=siteStore, apiName=apiName, apiKey=apiKey)
        existingKey.apiKey = apiKey
        return existingKey
    setKeyForAPI = classmethod(setKeyForAPI)



def _makeSiteConfiguration(currentVersion, oldSite, couldHavePorts):
    from xmantissa.publicweb import AnonymousSite

    newSite = oldSite.upgradeVersion(
        'mantissa_web_powerup', currentVersion, 6,
        hitCount=oldSite.hitCount)

    if newSite.store.parent is not None:
        return newSite

    # SiteConfiguration dependsOn LoginSystem.  LoginSystem was probably
    # installed by the mantissa axiomatic command.  During the dependency
    # system conversion, that command was changed to use installOn on the
    # LoginSystem.  However, no upgrader was supplied to create the new
    # dependency state.  Consequently, there may be none.  Consequently, a new
    # LoginSystem will be created if an item which dependsOn LoginSystem is
    # installed.  This would be bad.  So, set up the necessary dependency state
    # here, before instantiating SiteConfiguration. -exarkun

    # Addendum: it is almost certainly the case that there are not legitimate
    # configurations which lack a LoginSystem.  However, a large number of
    # database upgrade tests construct unrealistic databases.  One aspect of
    # the unrealism is that they lack a LoginSystem.  Therefore, rather than
    # changing all the bogus stubs and regenerating the stubs, I will just
    # support the case where LoginSystem is missing.  However, note that a
    # LoginSystem upgrader may invalidate this check and result in a duplicate
    # being created anyway. -exarkun
    loginSystem = oldSite.store.findUnique(LoginSystem, default=None)
    if loginSystem is not None and installedOn(loginSystem) is None:
        installOn(loginSystem, oldSite.store)

    uninstallFrom(oldSite.store, oldSite)

    site = SiteConfiguration(
        store=oldSite.store,
        httpLog=oldSite.store.filesdir.child('httpd.log'),
        hostname=getattr(oldSite, "hostname", None) or u"localhost")
    installOn(site, site.store)

    anonymousAvatar = AnonymousSite(store=oldSite.store)
    installOn(anonymousAvatar, oldSite.store)

    if couldHavePorts:
        for tcp in site.store.query(TCPPort, TCPPort.factory == oldSite):
            tcp.factory = site
        for ssl in site.store.query(SSLPort, SSLPort.factory == oldSite):
            ssl.factory = site
    else:
        if oldSite.portNumber is not None:
            port = TCPPort(
                store=oldSite.store,
                portNumber=oldSite.portNumber,
                factory=site)
            installOn(port, oldSite.store)

        securePortNumber = oldSite.securePortNumber
        certificateFile = oldSite.certificateFile
        if securePortNumber is not None and certificateFile:
            oldCertPath = site.store.dbdir.preauthChild(certificateFile)
            newCertPath = site.store.newFilePath('server.pem')
            oldCertPath.moveTo(newCertPath)
            port = SSLPort(
                store=site.store,
                portNumber=oldSite.securePortNumber,
                certificatePath=newCertPath,
                factory=site)
            installOn(port, site.store)

    newSite.deleteFromStore()


declareLegacyItem(
    WebSite.typeName, 1, dict(
        hitCount = integer(default=0),
        installedOn = reference(),
        portNumber = integer(default=0),
        securePortNumber = integer(default=0),
        certificateFile = bytes(default=None)))

def upgradeWebSite1To6(oldSite):
    return _makeSiteConfiguration(1, oldSite, False)
upgrade.registerUpgrader(upgradeWebSite1To6, 'mantissa_web_powerup', 1, 6)

declareLegacyItem(
    WebSite.typeName, 2, dict(
        hitCount = integer(default=0),
        installedOn = reference(),
        portNumber = integer(default=0),
        securePortNumber = integer(default=0),
        certificateFile = bytes(default=None),
        httpLog = bytes(default=None)))

def upgradeWebSite2to6(oldSite):
    # This is dumb and we should have a way to run procedural upgraders.
    newSite = _makeSiteConfiguration(2, oldSite, False)
    staticMistake = newSite.store.findUnique(StaticSite,
                                             StaticSite.prefixURL == u'static/mantissa',
                                             default=None)
    if staticMistake is not None:
        # Ugh, need cascading deletes
        staticMistake.store.powerDown(staticMistake, ISessionlessSiteRootPlugin)
        staticMistake.deleteFromStore()

    return newSite
upgrade.registerUpgrader(upgradeWebSite2to6, 'mantissa_web_powerup', 2, 6)

declareLegacyItem(
    WebSite.typeName, 3, dict(
        hitCount = integer(default=0),
        installedOn = reference(),
        portNumber = integer(default=0),
        securePortNumber = integer(default=0),
        certificateFile = bytes(default=None),
        httpLog = bytes(default=None)))

def upgradeWebsite3to6(oldSite):
    return _makeSiteConfiguration(3, oldSite, False)
upgrade.registerUpgrader(upgradeWebsite3to6, 'mantissa_web_powerup', 3, 6)

declareLegacyItem(
    WebSite.typeName, 4, dict(
        hitCount=integer(default=0),
        installedOn=reference(),
        hostname=text(default=None),
        portNumber=integer(default=0),
        securePortNumber=integer(default=0),
        certificateFile=bytes(default=0),
        httpLog=bytes(default=None)))

def upgradeWebsite4to6(oldSite):
    return _makeSiteConfiguration(4, oldSite, False)
upgrade.registerUpgrader(upgradeWebsite4to6, 'mantissa_web_powerup', 4, 6)

declareLegacyItem(
    WebSite.typeName, 5, dict(
        hitCount=integer(default=0),
        installedOn=reference(),
        hostname=text(default=None),
        httpLog=bytes(default=None)))

def upgradeWebsite5to6(oldSite):
    """
    Create a L{SiteConfiguration} if this is a site store's L{WebSite}.
    """
    return _makeSiteConfiguration(5, oldSite, True)
upgrade.registerUpgrader(upgradeWebsite5to6, WebSite.typeName, 5, 6)




"""
Tests for L{xmantissa.website.WebSite}'s discovery of L{ISiteRootPlugin}
powerups.
"""

from twisted.trial import unittest

from nevow.testutil import FakeRequest

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import text
from axiom.dependency import installOn

from xmantissa.website import PrefixURLMixin, WebSite
from xmantissa.ixmantissa import ISiteRootPlugin

from zope.interface import implements


class Dummy:
    def __init__(self, pfx):
        self.pfx = pfx



class PrefixTester(Item, PrefixURLMixin):

    implements(ISiteRootPlugin)

    sessioned = True

    typeName = 'test_prefix_widget'
    schemaVersion = 1

    prefixURL = text()

    def createResource(self):
        return Dummy(self.prefixURL)

    def installSite(self):
        """
        Not using the dependency system for this class because multiple
        instances can be installed.
        """
        for iface, priority in self.__getPowerupInterfaces__([]):
            self.store.powerUp(self, iface, priority)



class SiteRootTest(unittest.TestCase):
    def test_prefixPriorityMath(self):
        """
        L{WebSite.locateChild} returns the most specific L{ISiteRootPlugin}
        based on I{prefixURL} and the request path segments.
        """
        store = Store()

        PrefixTester(store=store, prefixURL=u"hello").installSite()
        PrefixTester(store=store, prefixURL=u"").installSite()

        website = WebSite(store=store)
        installOn(website, store)

        res, segs = website.locateChild(FakeRequest(), ('hello',))
        self.assertEquals(res.pfx, 'hello')
        self.assertEquals(segs, ())

        res, segs = website.locateChild(FakeRequest(), ('',))
        self.assertEquals(res.pfx, '')
        self.assertEquals(segs, ('',))
